import itertools
import logging
from collections import namedtuple

from claim.models import ClaimItem, Claim, ClaimService, ClaimDedRem, ClaimDetail, ClaimServiceService, ClaimServiceItem
from core import utils
from core.datetimes.shared import datetimedelta
from core.utils import filter_validity
from django.db import connection
from django.db.models import Sum, Q
from django.db.models.functions import Coalesce
from django.utils.translation import gettext as _
from insuree.models import InsureePolicy
from medical.models import Service, ServiceService, ServiceItem
from medical_pricelist.models import ItemsPricelistDetail, ServicesPricelistDetail
from policy.models import Policy
from product.models import Product, ProductItem, ProductService, ProductItemOrService

from .apps import ClaimConfig
from .utils import get_queryset_valid_at_date

logger = logging.getLogger(__name__)

REJECTION_REASON_INVALID_ITEM_OR_SERVICE = 1
REJECTION_REASON_NOT_IN_PRICE_LIST = 2
REJECTION_REASON_NO_PRODUCT_FOUND = 3
REJECTION_REASON_CATEGORY_LIMITATION = 4
REJECTION_REASON_FREQUENCY_FAILURE = 5
# REJECTION_REASON_DUPLICATED = 6
REJECTION_REASON_FAMILY = 7
# REJECTION_REASON_ICD_NOT_IN_LIST = 8
REJECTION_REASON_TARGET_DATE = 9
REJECTION_REASON_CARE_TYPE = 10
REJECTION_REASON_MAX_HOSPITAL_ADMISSIONS = 11
REJECTION_REASON_MAX_VISITS = 12
REJECTION_REASON_MAX_CONSULTATIONS = 13
REJECTION_REASON_MAX_SURGERIES = 14
REJECTION_REASON_MAX_DELIVERIES = 15
REJECTION_REASON_QTY_OVER_LIMIT = 16
REJECTION_REASON_WAITING_PERIOD_FAIL = 17
REJECTION_REASON_MAX_ANTENATAL = 19
REJECTION_REASON_INVALID_CLAIM = 20

def validate_claim(claim, check_max):
    """
    Based on the legacy validation, this method returns standard codes along with details
    :param claim: claim to be verified
    :param check_max: max amount to validate. Everything above will be rejected
    :return: (result_code, error_details)
    """
    logger.debug(f"Validating claim {claim.uuid}")
    if ClaimConfig.default_validations_disabled:
        return []
    errors = []
    detail_errors = []
    errors += validate_target_date(claim)
    if len(errors) == 0:
        errors += validate_family(claim, claim.insuree)
    if len(errors) == 0:
        detail_errors += validate_claimitems(claim)
        detail_errors += validate_claimservices(claim)

    if check_max:
        # we went over the maximum for a category, all items and services in the claim are rejected
        over_category_errors = [
            x for x in detail_errors if x['code'] in [REJECTION_REASON_MAX_HOSPITAL_ADMISSIONS,
                                                      REJECTION_REASON_MAX_VISITS,
                                                      REJECTION_REASON_MAX_CONSULTATIONS,
                                                      REJECTION_REASON_MAX_SURGERIES,
                                                      REJECTION_REASON_MAX_DELIVERIES,
                                                      REJECTION_REASON_MAX_ANTENATAL]]
        if len(over_category_errors) > 0:
            rtn_items_rejected = claim.items.filter(validity_to__isnull=True) \
                .update(status=ClaimItem.STATUS_REJECTED,
                        qty_approved=0,
                        rejection_reason=over_category_errors[0]['code'])
            rtn_services_rejected = claim.services.filter(validity_to__isnull=True) \
                .update(status=ClaimService.STATUS_REJECTED,
                        qty_approved=0,
                        rejection_reason=over_category_errors[0]['code'])
        else:
            rtn_items_rejected = claim.items.filter(validity_to__isnull=True) \
                .exclude(rejection_reason=0).exclude(rejection_reason__isnull=True) \
                .update(status=ClaimItem.STATUS_REJECTED, qty_approved=0)
            rtn_services_rejected = claim.services.filter(validity_to__isnull=True) \
                .exclude(rejection_reason=0).exclude(rejection_reason__isnull=True) \
                .update(status=ClaimService.STATUS_REJECTED, qty_approved=0)
        if rtn_items_rejected or rtn_services_rejected:
            logger.debug(f"Marked {rtn_items_rejected} items as rejected and {rtn_services_rejected} services")

    rtn_items_passed = claim.items.filter(validity_to__isnull=True) \
        .exclude(status=ClaimItem.STATUS_REJECTED) \
        .update(status=ClaimItem.STATUS_PASSED)
    rtn_services_passed = claim.services.filter(validity_to__isnull=True) \
        .exclude(status=ClaimService.STATUS_REJECTED) \
        .update(status=ClaimService.STATUS_PASSED)

    if rtn_items_passed + rtn_services_passed == 0:
        errors += [{'code': REJECTION_REASON_INVALID_ITEM_OR_SERVICE,
                    'message': _("claim.validation.all_items_and_services_rejected") % {
                        'code': claim.code},
                    'detail': claim.uuid}]
        if len(detail_errors)>0:
            errors += detail_errors
        claim.status = Claim.STATUS_REJECTED
        claim.rejection_reason = REJECTION_REASON_INVALID_ITEM_OR_SERVICE
        claim.save()
    logger.debug(f"Validation found {len(errors)} error(s)")
    return errors


def validate_claimitems(claim):
    errors = []
    target_date = claim.date_from if claim.date_from else claim.date_to
    for claimitem in claim.items.all():
        if not claimitem.rejection_reason:
            errors += validate_claimitem_validity(claim, claimitem)
            if not claimitem.rejection_reason:
                errors += validate_claimitem_in_price_list(claim, claimitem)
            if not claimitem.rejection_reason:
                errors += validate_claimdetail_care_type(claim, claimitem)
            if not claimitem.rejection_reason:
                errors += validate_claimdetail_limitation_fail(claim, claimitem)
            if not claimitem.rejection_reason:
                errors += validate_claimitem_frequency(claim, claimitem)
            if not claimitem.rejection_reason:
                errors += validate_item_product_family(
                    claimitem=claimitem,
                    target_date=target_date,
                    item=claimitem.item,
                    insuree_id=claim.insuree_id,
                    adult=claim.insuree.is_adult(target_date)
                )
            if claimitem.rejection_reason:
                claimitem.status = ClaimItem.STATUS_REJECTED
            else:
                claimitem.rejection_reason = 0
                claimitem.status = ClaimItem.STATUS_PASSED
            claimitem.save()
    return errors


def validate_claimservices(claim):
    errors = []
    target_date = claim.date_from if claim.date_from else claim.date_to
    base_category = get_claim_category(claim)

    for claimservice in claim.services.all():
        if not claimservice.rejection_reason:
            errors += validate_claimservice_validity(claim, claimservice)
            if not claimservice.rejection_reason:
                errors += validate_claimservice_in_price_list(claim, claimservice)
            if not claimservice.rejection_reason:
                errors += validate_claimdetail_care_type(claim, claimservice)
            if not claimservice.rejection_reason:
                errors += validate_claimservice_frequency(claim, claimservice)
            if not claimservice.rejection_reason:
                errors += validate_claimdetail_limitation_fail(claim, claimservice)
            if not claimservice.rejection_reason:
                errors += validate_service_product_family(
                    claimservice=claimservice,
                    target_date=target_date,
                    service=claimservice.service,
                    insuree_id=claim.insuree_id,
                    adult=claim.insuree.is_adult(target_date),
                    base_category=base_category,
                    claim=claim,
                )
            if claimservice.rejection_reason:
                claimservice.status = ClaimService.STATUS_REJECTED
            else:
                claimservice.rejection_reason = 0
                claimservice.status = ClaimService.STATUS_PASSED
            claimservice.save()
    return errors


def validate_claimitem_validity(claim, claimitem):
    # In the stored procedure, this check used a complex query to get the latest item but the latest item seems to
    # always be updated.
    # select *
    # from tblClaimItems tCI inner join tblItems tI on tCI.ItemID = tI.ItemID
    # where ti.ValidityTo is not null and tI.LegacyID is not null;
    # gives no result, so no claimitem is pointing to an old item and the complex query always fetched the last one.
    # Here, claimitem.item.legacy_id is always None
    errors = []
    if claimitem.validity_to is None and claimitem.item.validity_to is not None:
        claimitem.rejection_reason = REJECTION_REASON_INVALID_ITEM_OR_SERVICE
        errors += [{'code': REJECTION_REASON_INVALID_ITEM_OR_SERVICE,
                    'message': _("claim.validation.claimitem_validity") % {
                        'code': claim.code
                    },
                    'detail': claim.uuid}]
    return errors


def validate_claimservice_validity(claim, claimservice):
    # See note in validate_claimitem_validity
    errors = []
    if claimservice.validity_to is None and claimservice.service.validity_to is not None:
        claimservice.rejection_reason = REJECTION_REASON_INVALID_ITEM_OR_SERVICE
        errors += [{'code': REJECTION_REASON_INVALID_ITEM_OR_SERVICE,
                    'message': _("claim.validation.claimservice_validity") % {
                        'code': claim.code
                    },
                    'detail': claim.uuid}]
    return errors


def __get_claim_target_date(claim):
    return claim.date_to if claim.date_to else claim.date_from


def validate_claimitem_in_price_list(claim, claimitem):
    errors = []
    target_date = __get_claim_target_date(claim)
    pricelist_detail_qs = ItemsPricelistDetail.objects \
        .filter(item_id=claimitem.item_id,
                validity_to__isnull=True,
                items_pricelist=claim.health_facility.items_pricelist,
                items_pricelist__validity_to__isnull=True
                )
    pricelist_detail = get_queryset_valid_at_date(pricelist_detail_qs, target_date).first()
    if not pricelist_detail:
        claimitem.rejection_reason = REJECTION_REASON_NOT_IN_PRICE_LIST
        errors += [{'code': REJECTION_REASON_NOT_IN_PRICE_LIST,
                    'message': _("claim.validation.claimitem_in_price_list_validity") % {
                        'code': claim.code
                    },
                    'detail': claim.uuid}]
    return errors


def validate_claimservice_in_price_list(claim, claimservice):
    errors = []
    target_date = __get_claim_target_date(claim)
    pricelist_detail_qs = ServicesPricelistDetail.objects \
        .filter(service_id=claimservice.service_id,
                services_pricelist=claim.health_facility.services_pricelist,
                services_pricelist__validity_to__isnull=True
                )
    pricelist_detail = get_queryset_valid_at_date(pricelist_detail_qs, target_date).first()
    if not pricelist_detail:
        claimservice.rejection_reason = REJECTION_REASON_NOT_IN_PRICE_LIST
        errors += [{'code': REJECTION_REASON_NOT_IN_PRICE_LIST,
                    'message': _("claim.validation.claimservice_in_price_list_validity") % {
                        'code': claim.code
                    },
                    'detail': claim.uuid}]
    return errors


def validate_claimdetail_care_type(claim, claimdetail):
    errors = []
    care_type = claimdetail.itemsvc.care_type
    hf_care_type = claim.health_facility.care_type if claim.health_facility.care_type else 'B'
    target_date = __get_claim_target_date(claim)

    if (
            care_type == 'I' and (
            hf_care_type == 'O'
            or target_date == claim.date_from)
    ) or (
            care_type == 'O' and (
            hf_care_type == 'I'
            or target_date != claim.date_from)
    ):
        claimdetail.rejection_reason = REJECTION_REASON_CARE_TYPE
        errors += [{'code': REJECTION_REASON_CARE_TYPE,
                    'message': _("claim.validation.claimdetail_care_type_validity") % {
                        'code': claim.code
                    },
                    'detail': claim.uuid}]
    return errors


def validate_claimdetail_limitation_fail(claim, claimdetail):
    # if the mask is empty, it should be valid for everyone
    if claimdetail.itemsvc.patient_category == 0:
        return []
    errors = []
    target_date = __get_claim_target_date(claim)
    patient_category_mask = utils.patient_category_mask(
        claim.insuree, target_date)
    
    if claimdetail.itemsvc.patient_category & patient_category_mask != patient_category_mask:
        claimdetail.rejection_reason = REJECTION_REASON_CATEGORY_LIMITATION
        errors += [{'code': REJECTION_REASON_CATEGORY_LIMITATION,
                    'message': _("claim.validation.claimdetail_limitation_validity") % {
                        'code': claim.code
                    },
                    'detail': claim.uuid}]
    return errors


def frequency_check(qs, claim, elt):
    td = claim.date_from if not claim.date_to else claim.date_to
    delta = datetimedelta(days=elt.frequency)
    return qs \
        .annotate(target_date=Coalesce("claim__date_to", "claim__date_from")) \
        .filter(Q(rejection_reason=0) | Q(rejection_reason__isnull=True),
                validity_to__isnull=True,
                target_date__gte=td - delta,
                status=ClaimDetail.STATUS_PASSED,
                claim__insuree_id=claim.insuree_id,
                claim__status__gt=Claim.STATUS_ENTERED
                ) \
        .exclude(claim__uuid=claim.uuid) \
        .exists()


def validate_claimitem_frequency(claim, claimitem):
    errors = []
    if claimitem.item.frequency and \
            frequency_check(ClaimItem.objects.filter(item=claimitem.item), claim, claimitem.item):
        claimitem.rejection_reason = REJECTION_REASON_FREQUENCY_FAILURE
        errors += [{'code': REJECTION_REASON_FREQUENCY_FAILURE,
                    'message': _("claim.validation.claimitem_frequency_validity") % {
                        'code': claim.code
                    },
                    'detail': claim.uuid}]
    return errors


def validate_claimservice_frequency(claim, claimservice):
    errors = []
    if claimservice.service.frequency and \
            frequency_check(ClaimService.objects.filter(service=claimservice.service), claim, claimservice.service):
        claimservice.rejection_reason = REJECTION_REASON_FREQUENCY_FAILURE
        errors += [{'code': REJECTION_REASON_FREQUENCY_FAILURE,
                    'message': _("claim.validation.claimservice_frequency_validity") % {
                        'code': claim.code
                    },
                    'detail': claim.uuid}]
    return errors


def validate_target_date(claim):
    errors = []
    if (claim.date_from is None and claim.date_to is None) \
            or claim.date_claimed < claim.date_from:
        claim.reject(REJECTION_REASON_TARGET_DATE)
        errors += [{'code': REJECTION_REASON_TARGET_DATE,
                    'message': _("claim.validation.target_date") % {
                        'code': claim.code
                    },
                    'detail': claim.uuid}]
    return errors


def validate_family(claim, insuree):
    errors = []
    if insuree.validity_to is not None:
        errors += [{'code': REJECTION_REASON_FAMILY,
                    'message': _("claim.validation.family.insuree_validity") % {
                        'code': claim.code,
                        'insuree': str(insuree)},
                    'detail': claim.uuid}]
    elif insuree.family is None:
        errors += [{'code': REJECTION_REASON_FAMILY,
                    'message': _("claim.validation.family.no_family") % {
                        'code': claim.code,
                        'insuree': str(insuree)},
                    'detail': claim.uuid}]
    elif insuree.family.validity_to is not None:
        errors += [{'code': REJECTION_REASON_FAMILY,
                    'message': _("claim.validation.family.family_validity") % {
                        'code': claim.code,
                        'insuree': str(insuree)},
                    'detail': claim.uuid}]
    if len(errors) > 0:
        claim.reject(REJECTION_REASON_FAMILY)
    return errors


def validate_item_product_family(claimitem, target_date, item, insuree_id, adult):
    errors = []
    found = False
    with get_products(target_date, item.id, insuree_id, adult, 'Item') as cursor:
        for (product_id, product_item_id, insuree_policy_effective_date, policy_effective_date, expiry_date,
             policy_stage) in cursor.fetchall():
            found = True
            core = __import__("core")
            insuree_policy_effective_date = core.datetime.date.from_ad_date(
                insuree_policy_effective_date)
            expiry_date = core.datetime.date.from_ad_date(expiry_date)
            product_item = ProductItem.objects.get(pk=product_item_id)
            # START CHECK 17 --> Item/Service waiting period violation (17)
            errors = check_service_item_waiting_period(policy_stage, policy_effective_date,
                                                       insuree_policy_effective_date,
                                                       item, adult, product_item, target_date, claimitem)

            # **** START CHECK 16 --> Item/Service Maximum provision (16)*****
            errors += check_service_item_max_provision(adult, product_item, item, insuree_policy_effective_date,
                                                       expiry_date, insuree_id, claimitem)
        if not found:
            claimitem.rejection_reason = REJECTION_REASON_NO_PRODUCT_FOUND
            errors += [{'code': REJECTION_REASON_NO_PRODUCT_FOUND,
                        'message': _("claim.validation.product_family.no_product_found") % {
                            'code': claimitem.claim.code,
                            'element': str(item)},
                        'detail': claimitem.claim.uuid}]

    return errors


# noinspection DuplicatedCode
def validate_service_product_family(claimservice, target_date, service, insuree_id, adult, base_category, claim):
    errors = []
    found = False
    with get_products(target_date, service.id, insuree_id, adult, 'Service') as cursor:
        for (product_id, product_service_id, insuree_policy_effective_date, policy_effective_date, expiry_date,
             policy_stage) in cursor.fetchall():
            found = True
            core = __import__("core")
            insuree_policy_effective_date = core.datetime.date.from_ad_date(
                insuree_policy_effective_date)
            expiry_date = core.datetime.date.from_ad_date(expiry_date)
            product_service = ProductService.objects.get(pk=product_service_id)

            # START CHECK 17 --> Item/Service waiting period violation (17)
            errors += check_service_item_waiting_period(policy_stage, policy_effective_date,
                                                        insuree_policy_effective_date, service, adult,
                                                        product_service, target_date, claimservice)

            # **** START CHECK 16 --> Item/Service Maximum provision (16)*****
            errors += check_service_item_max_provision(adult, product_service, service, insuree_policy_effective_date,
                                                       expiry_date, insuree_id, claimservice)

            # Each violation is meant to interrupt the validation
            error_len = len(errors)
            product = Product.objects.filter(pk=product_id).first()
            if base_category != 'O':
                errors += check_claim_max_no_category(base_category, product, expiry_date, insuree_id,
                                                      insuree_policy_effective_date, claim, claimservice)
                if error_len != len(errors):
                    break

        if not found:
            claimservice.rejection_reason = REJECTION_REASON_NO_PRODUCT_FOUND
            errors += [{'code': REJECTION_REASON_NO_PRODUCT_FOUND,
                        'message': _("claim.validation.product_family.no_product_found") % {
                            'code': claimservice.claim.code,
                            'element': str(service)},
                        'detail': claimservice.claim.uuid}]

    return errors

def check_service_item_waiting_period(policy_stage, policy_effective_date, insuree_policy_effective_date, service_or_item,
                                 adult, product_service_item, target_date, claim_service_item):
    errors = []
    waiting_period = None
    if policy_stage == 'N' or policy_effective_date < insuree_policy_effective_date:
        if adult:
            waiting_period = product_service_item.waiting_period_adult
        else:
            waiting_period = product_service_item.waiting_period_child
    if waiting_period and target_date < \
            (insuree_policy_effective_date.to_datetime() + datetimedelta(months=waiting_period)):
        claim_service_item.rejection_reason = REJECTION_REASON_WAITING_PERIOD_FAIL
        errors += [{'code': REJECTION_REASON_WAITING_PERIOD_FAIL,
                    'message': _("claim.validation.product_family.waiting_period") % {
                        'code': claim_service_item.claim.code,
                        'element': str(service_or_item)},
                    'detail': claim_service_item.claim.uuid}]
    return errors


def check_service_item_max_provision(adult, product_service_item, service_or_item, insuree_policy_effective_date,
                                     expiry_date, insuree_id, claim_service_item):
    errors = []
    if adult:
        limit_no = product_service_item.limit_no_adult
    else:
        limit_no = product_service_item.limit_no_child
    if limit_no is not None and limit_no >= 0:
        # count qty provided
        total_qty_provided = _get_total_qty_provided(claim_service_item, service_or_item, insuree_policy_effective_date,
                                                     expiry_date, insuree_id)
        qty = total_qty_provided + claim_service_item.qty_provided if claim_service_item.qty_approved is None \
                                                                   else claim_service_item.qty_approved
        if qty > limit_no:
            # it would be good to add a warning msg, here is a related ticket: OTC-943
            if total_qty_provided < limit_no:
                remaining_qty = limit_no - total_qty_provided
                if claim_service_item.qty_approved is None:
                    claim_service_item.qty_provided = remaining_qty
                else:
                    claim_service_item.qty_approved = remaining_qty
                claim_service_item.save()
            else:
                claim_service_item.rejection_reason = REJECTION_REASON_QTY_OVER_LIMIT
                errors += [{'code': REJECTION_REASON_QTY_OVER_LIMIT,
                            'message': _("claim.validation.product_family.max_nb_allowed") % {
                                'code': claim_service_item.claim.code,
                                'element': str(service_or_item),
                                'provided': total_qty_provided,
                                'max': limit_no},
                            'detail': claim_service_item.claim.uuid}]

    return errors


def _get_total_qty_provided(claim_service_item, service_or_item, insuree_policy_effective_date,
                            expiry_date, insuree_id):
    return claim_service_item.__class__.objects \
            .annotate(target_date=Coalesce("claim__date_to", "claim__date_from")) \
            .filter(Q(rejection_reason=0) | Q(rejection_reason__isnull=True),
                    validity_to__isnull=True,
                    **{
                        f"{'service' if isinstance(service_or_item, Service) else 'item'}_id": service_or_item.id},
                    policy__validity_to__isnull=True,
                    target_date__gte=insuree_policy_effective_date,
                    target_date__lte=expiry_date,
                    claim__insuree_id=insuree_id,
                    claim__status__gt=Claim.STATUS_ENTERED,
                    claim__validity_to__isnull=True
                    ) \
            .aggregate(total_qty_provided=Sum(Coalesce("qty_approved", "qty_provided"))) \
            .get("total_qty_provided") or 0

def check_claim_max_no_category(base_category, product, expiry_date, insuree_id,
                                insuree_policy_effective_date, claim, claimservice):
    errors = []
    category_dict = {
        'C': {'max': product.max_no_consultation,
              "reason": REJECTION_REASON_MAX_CONSULTATIONS,
              "message": "claim.validation.product_family.max_nb_consultation"},
        'S': {"max": product.max_no_surgery,
              "reason": REJECTION_REASON_MAX_SURGERIES,
              "message": "claim.validation.product_family.max_nb_surgeries"},
        'D': {"max": product.max_no_delivery,
              "reason": REJECTION_REASON_MAX_DELIVERIES,
              "message": "claim.validation.product_family.max_nb_deliveries"},
        'A': {"max": product.max_no_antenatal,
              "reason": REJECTION_REASON_MAX_ANTENATAL,
              "message": "claim.validation.product_family.max_nb_antenatal"},
        'H': {"max": product.max_no_hospitalization,
              "reason": REJECTION_REASON_MAX_HOSPITAL_ADMISSIONS,
              "message": "claim.validation.product_family.max_nb_hospitalizations"},
        'V': {"max": product.max_no_visits,
              "reason": REJECTION_REASON_MAX_VISITS,
              "message": "claim.validation.product_family.max_nb_visits"},
    }.get(base_category)

    if category_dict['max'] is not None and category_dict['max'] >= 0:
        count = get_claim_queryset_by_category(expiry_date, insuree_id, insuree_policy_effective_date, base_category, claim) \
            .count()
        if count and count >= category_dict['max']:
            claimservice.rejection_reason = category_dict['reason']
            errors += [{'code': category_dict['reason'],
                        'message': _(category_dict['message']) % {
                            'code': claimservice.claim.code,
                            'count': count,
                            'max': category_dict['max']},
                        'detail': claimservice.claim.uuid}]
    return errors


def get_claim_queryset_by_category(expiry_date, insuree_id, insuree_policy_effective_date, category, claim=None):
    queryset = Claim.objects \
        .annotate(target_date=Coalesce("date_to", "date_from")) \
        .filter(insuree_id=insuree_id,
                validity_to__isnull=True,
                status__gt=Claim.STATUS_ENTERED,
                target_date__gte=insuree_policy_effective_date,
                target_date__lte=expiry_date)
    if claim:
        queryset = queryset.exclude(uuid=claim.uuid)
    if category == 'V':
        queryset = queryset.filter(
            Q(category=category) | Q(category__isnull=True))
    else:
        queryset = queryset.filter(category=category)
    return queryset


def get_products(target_date, elt_id, insuree_id, adult, item_or_service):
    cursor = connection.cursor()
    waiting_period = "WaitingPeriodAdult" if adult else "WaitingPeriodChild"
    # about deductions and ceilings...
    # tblProduct.DedInsuree, tblProduct.DedOPInsuree, tblProduct.DedIPInsuree,
    # tblProduct.MaxInsuree, tblProduct.MaxOPInsuree, tblProduct.MaxIPInsuree,
    # tblProduct.DedTreatment, tblProduct.DedOPTreatment, tblProduct.DedIPTreatment,
    # tblProduct.MaxTreatment, tblProduct.MaxOPTreatment, tblProduct.MaxIPTreatment,
    # tblProduct.DedPolicy, tblProduct.DedOPPolicy, tblProduct.DedIPPolicy,
    # tblProduct.MaxPolicy, tblProduct.MaxOPPolicy, tblProduct.MaxIPPolicy
    if connection.vendor == "postgresql":
        sql = f"""
                    SELECT 
                        "tblProduct"."ProdID", "tblProduct{item_or_service}s"."Prod{item_or_service}ID",            
                        "tblInsureePolicy"."EffectiveDate",
                        "tblPolicy"."EffectiveDate",
                        CASE
                            WHEN "tblPolicy"."ExpiryDate" < "tblInsureePolicy"."ExpiryDate" THEN "tblPolicy"."ExpiryDate"
                            ELSE "tblInsureePolicy"."ExpiryDate"
                        END AS "ExpiryDate",
                        "tblPolicy"."PolicyStage"
                    FROM "tblInsuree" 
                        INNER JOIN "tblInsureePolicy" ON "tblInsureePolicy"."InsureeID" = "tblInsuree"."InsureeID"
                        LEFT OUTER JOIN "tblPolicy"
                        LEFT OUTER JOIN "tblProduct" ON "tblPolicy"."ProdID" = "tblProduct"."ProdID"
                        INNER JOIN "tblProduct{item_or_service}s" 
                            ON "tblProduct"."ProdID" = "tblProduct{item_or_service}s"."ProdID"            
                        RIGHT OUTER JOIN "tblFamilies" ON "tblPolicy"."FamilyID" = "tblFamilies"."FamilyID"
                        ON "tblInsuree"."FamilyID" = "tblFamilies"."FamilyID"
                    WHERE ("tblInsuree"."ValidityTo" IS NULL) AND ("tblInsuree"."InsureeID" = %s)
                        AND ("tblInsureePolicy"."PolicyId" = "tblPolicy"."PolicyID")
                        AND ("tblPolicy"."ValidityTo" IS NULL)
                        AND ("tblPolicy"."EffectiveDate" <= %s) AND ("tblPolicy"."ExpiryDate" >= %s)
                        AND ("tblInsureePolicy"."ValidityTo" IS NULL)
                        AND ("tblInsureePolicy"."EffectiveDate" <= %s) AND ("tblInsureePolicy"."ExpiryDate" >= %s)
                        AND ("tblPolicy"."PolicyStatus" in ({Policy.STATUS_ACTIVE}, {Policy.STATUS_EXPIRED}))
                        AND ("tblProduct{item_or_service}s"."ValidityTo" IS NULL) 
                            AND ("tblProduct{item_or_service}s"."{item_or_service}ID" = %s)
                    ORDER BY "tblPolicy"."EffectiveDate" + coalesce("tblProduct{item_or_service}s"."{waiting_period}", 0) 
                        * INTERVAL '1 MONTH'            
                """
    else:
        sql = f"""
            SELECT 
                tblProduct.ProdID, tblProduct{item_or_service}s.Prod{item_or_service}ID,            
                tblInsureePolicy.EffectiveDate,
                tblPolicy.EffectiveDate,
                CASE
                    WHEN tblPolicy.ExpiryDate < tblInsureePolicy.ExpiryDate THEN tblPolicy.ExpiryDate
                    ELSE tblInsureePolicy.ExpiryDate
                END AS ExpiryDate,
                tblPolicy.PolicyStage
            FROM tblInsuree 
                INNER JOIN tblInsureePolicy ON tblInsureePolicy.InsureeID = tblInsuree.InsureeID
                LEFT OUTER JOIN tblPolicy
                LEFT OUTER JOIN  tblProduct ON tblPolicy.ProdID = tblProduct.ProdID
                INNER JOIN tblProduct{item_or_service}s ON tblProduct.ProdID = tblProduct{item_or_service}s.ProdID            
                RIGHT OUTER JOIN tblFamilies ON tblPolicy.FamilyID = tblFamilies.FamilyID
                ON tblInsuree.FamilyID = tblFamilies.FamilyID
            WHERE (tblInsuree.ValidityTo IS NULL) AND (tblInsuree.InsureeId = %s)
                AND (tblInsureePolicy.PolicyId = tblPolicy.PolicyID)
                AND (tblPolicy.ValidityTo IS NULL)
                AND (tblPolicy.EffectiveDate <= %s) AND (tblPolicy.ExpiryDate >= %s)
                AND (tblInsureePolicy.ValidityTo IS NULL)
                AND (tblInsureePolicy.EffectiveDate <= %s) AND (tblInsureePolicy.ExpiryDate >= %s)
                AND (tblPolicy.PolicyStatus in ({Policy.STATUS_ACTIVE}, {Policy.STATUS_EXPIRED}))
                AND (tblProduct{item_or_service}s.ValidityTo IS NULL) AND (tblProduct{item_or_service}s.{item_or_service}ID = %s)
            ORDER BY DATEADD(m,ISNULL(tblProduct{item_or_service}s.{waiting_period}, 0),
                tblPolicy.EffectiveDate)            
        """
    cursor.execute(sql,
                   [insuree_id, target_date, target_date, target_date, target_date, elt_id])
    return cursor


def get_claim_category(claim):
    """
    Determine the claim category based on its services:
    S = Surgery
    D = Delivery
    A = Antenatal care
    H = Hospitalization
    C = Consultation
    O = Other
    V = Visit
    :param claim: claim for which category should be determined
    :return: category if a service is defined, None if not service at all
    """

    service_categories = [
        Service.CATEGORY_SURGERY,
        Service.CATEGORY_DELIVERY,
        Service.CATEGORY_ANTENATAL,
        Service.CATEGORY_HOSPITALIZATION,
        Service.CATEGORY_CONSULTATION,
        Service.CATEGORY_OTHER,
        Service.CATEGORY_VISIT,
    ]
    target_date = __get_claim_target_date(claim)
    services = claim.services \
        .filter(validity_to__isnull=True, service__validity_to__isnull=True) \
        .values("service__category").distinct()
    claim_service_categories = [
        service["service__category"]
        for service in services
    ]
    if claim.date_from != target_date:
        claim_service_categories.append(Service.CATEGORY_HOSPITALIZATION)
    for category in service_categories:
        if category in claim_service_categories:
            claim_category = category
            break
    else:
        # One might expect "O" here but the legacy code uses "V"
        claim_category = Service.CATEGORY_VISIT

    return claim_category


def validate_assign_prod_elt(claim, elt, elt_ref, elt_qs):
    """
    This method checks the limits for the family and the insuree, child or adult for their limits
    between the copay percentage and fixed limit.
    :return: List of errors, only if no product could be found for now
    """
    visit_type_field = {
        "O": ("limitation_type", "limit_adult", "limit_child"),
        "E": ("limitation_type_e", "limit_adult_e", "limit_child_e"),
        "R": ("limitation_type_r", "limit_adult_r", "limit_child_r"),
    }
    logger.debug("[claim: %s] Assigning product for %s %s", claim.uuid, type(elt), elt.id)
    target_date = __get_claim_target_date(claim)
    visit_type = claim.visit_type if claim.visit_type and claim.visit_type in visit_type_field else "O"
    adult = claim.insuree.is_adult(target_date)
    (limitation_type_field, limit_adult, limit_child) = visit_type_field[visit_type]
    if elt.price_asked \
            and elt.price_approved \
            and elt.price_asked > elt.price_approved:
        claim_price = elt.price_asked
    else:
        claim_price = elt.price_approved
    logger.debug("[claim: %s] claim_price: %s", claim.uuid, claim_price)
    logger.debug("[claim: %s] Checking product itemsvc limit at date %s for family %s with field %s C for adult: %s",
                 claim.uuid, target_date, claim.insuree.family_id, limitation_type_field, adult)
    product_elt_c = _query_product_item_service_limit(
        target_date, claim.insuree.family_id, elt_qs, limitation_type_field, "C",
        limit_adult if adult else limit_child
    )
    logger.debug("[claim: %s] C product found: %s, checking product itemsvc limit at date %s for family %s "
                 "with field %s F for adult: %s", claim.uuid, product_elt_c is not None, target_date,
                 claim.insuree.family_id, limitation_type_field, adult)
    product_elt_f = _query_product_item_service_limit(
        target_date, claim.insuree.family_id, elt_qs, limitation_type_field, "F",
        limit_adult if adult else limit_child
    )
    logger.debug("[claim: %s] F found: %s", claim.uuid, product_elt_f is not None)
    if not product_elt_c and not product_elt_f:
        elt.rejection_reason = REJECTION_REASON_NO_PRODUCT_FOUND
        elt.save()
        return [{'code': REJECTION_REASON_NO_PRODUCT_FOUND,
                 'message': _("claim.validation.assign_prod.elt.no_product_code") % {
                     'code': claim.code,
                     'elt': str(elt_ref)},
                 'detail': claim.uuid}]

    if product_elt_f:
        fixed_limit = getattr(
            product_elt_f, limit_adult if adult else limit_child)
        logger.debug("[claim: %s] fixed_limit: %s", claim.uuid, fixed_limit)
    else:
        fixed_limit = None

    if product_elt_c:
        co_sharing_percent = getattr(
            product_elt_c, limit_adult if adult else limit_child)
        logger.debug("[claim: %s] co_sharing_percent: %s", claim.uuid, co_sharing_percent)
    else:
        co_sharing_percent = None

    # if both products exist, find the best one to use
    if product_elt_c and product_elt_f:
        if fixed_limit == 0 or fixed_limit > claim_price:
            product_elt = product_elt_f
            product_elt_c = None  # used in condition below
        else:
            if 100 - co_sharing_percent > 0:
                product_amount_own_f = claim_price - fixed_limit
                product_amount_own_c = (
                                               1 - co_sharing_percent / 100) * claim_price
                if product_amount_own_c > product_amount_own_f:
                    product_elt = product_elt_f
                    product_elt_c = None  # used in condition below
                else:
                    product_elt = product_elt_c
            else:
                product_elt = product_elt_c
    else:
        if product_elt_c:
            product_elt = product_elt_c
        else:
            product_elt = product_elt_f
            product_elt_c = None

    if product_elt is None:
        logger.warning(f"Could not find a suitable product from {type(elt)} {elt.id}")
    if product_elt.product_id is None:
        logger.warning(f"Found a productItem/Service for {type(elt)} {elt.id} but it does not have a product")
    logger.debug("[claim: %s] product_id found: %s", claim.uuid, product_elt.product_id)
    elt.product_id = product_elt.product_id
    logger.debug("[claim: %s] fetching policy for family %s", claim.uuid, claim.insuree.family_id)
    elt.policy = product_elt \
        .product \
        .policies.filter(
        family_id=claim.insuree.family_id,
        validity_to__isnull=True,
        effective_date__lte=target_date,
        expiry_date__gte=target_date,
        status__in=[Policy.STATUS_ACTIVE, Policy.STATUS_EXPIRED]
    ).first()
    if elt.policy is None:
        logger.warning(f"{type(elt)} id {elt.id} doesn't seem to have a valid policy with product"
                       f" {product_elt.product_id}")
    logger.debug("[claim: %s] setting policy %s", claim.uuid, elt.policy.id if elt.policy else None)
    elt.price_origin = product_elt.price_origin
    # The original code also sets claimservice.price_adjusted but it also always NULL
    if product_elt_c:
        elt.limitation = "C"
        elt.limitation_value = co_sharing_percent
    else:
        elt.limitation = "F"
        elt.limitation_value = fixed_limit
    logger.debug("[claim: %s] setting limitation %s to %s", claim.uuid, elt.limitation, elt.limitation_value)
    elt.save()
    return []


def validate_assign_prod_to_claimitems_and_services(claim):
    errors = []
    logger.debug("[claim: %s] validate_assign_prod_to_claimitems_and_services", claim.uuid)
    for claimitem in claim.items.filter(validity_to__isnull=True) \
            .filter(Q(rejection_reason=0) | Q(rejection_reason__isnull=True)):
        logger.debug("[claim: %s] validating item %s", claim.uuid, claimitem.id)
        errors += validate_assign_prod_elt(
            claim, claimitem, claimitem.item,
            ProductItem.objects.filter(item_id=claimitem.item_id))

    for claimservice in claim.services.filter(validity_to__isnull=True) \
            .filter(Q(rejection_reason=0) | Q(rejection_reason__isnull=True)):
        logger.debug("[claim: %s] validating service %s", claim.uuid, claimservice.id)
        errors += validate_assign_prod_elt(
            claim, claimservice, claimservice.service,
            ProductService.objects.filter(service_id=claimservice.service_id))

    logger.debug("[claim: %s] validate_assign_prod_to_claimitems_and_services nb of errors %s", claim.uuid, len(errors))
    return errors


def approved_amount(claim):
    app_item_value = claim.items \
        .annotate(value=Coalesce("qty_approved", "qty_provided") * Coalesce("price_approved", "price_asked")) \
        .filter(validity_to__isnull=True, status=ClaimItem.STATUS_PASSED) \
        .aggregate(Sum("value"))
    app_service_value = claim.services \
        .annotate(value=Coalesce("qty_approved", "qty_provided") * Coalesce("price_approved", "price_asked")) \
        .filter(validity_to__isnull=True, status=ClaimService.STATUS_PASSED) \
        .aggregate(Sum("value"))
    return (app_item_value['value__sum'] if app_item_value['value__sum'] else 0) + \
           (app_service_value['value__sum']
            if app_service_value['value__sum'] else 0)


def _query_product_item_service_limit(target_date, family_id, elt_qs,
                                      limitation_field, limitation_type,
                                      limit_ordering):
    return elt_qs \
        .filter(validity_to__isnull=True,
                product__validity_to__isnull=True,
                product__policies__family_id=family_id,
                product__policies__effective_date__lte=target_date,
                product__policies__expiry_date__gte=target_date,
                product__policies__validity_to__isnull=True,
                product__policies__status__in=[Policy.STATUS_ACTIVE, Policy.STATUS_EXPIRED],
                **{limitation_field: limitation_type}
                ) \
        .order_by("-" + limit_ordering) \
        .first()


Deductible = namedtuple('Deductible', ['amount', 'type', 'prev'])


def _get_dedrem(prefix, dedrem_type, field, product, claim, policy_id):
    if getattr(product, prefix + "_treatment", None):
        return Deductible(
            getattr(product, prefix + "_treatment", None),
            dedrem_type,
            0
        )
    if getattr(product, prefix + "_insuree", None):
        prev = ClaimDedRem.objects \
            .filter(policy_id=policy_id, insuree_id=claim.insuree_id) \
            .exclude(claim_id=claim.id) \
            .aggregate(sum=Sum(field))["sum"]
        return Deductible(
            getattr(product, prefix + "_insuree", None),
            dedrem_type,
            prev if prev else 0
        )
    if getattr(product, prefix + "_policy", None):
        prev = ClaimDedRem.objects \
            .filter(policy_id=policy_id) \
            .exclude(claim_id=claim.id) \
            .aggregate(sum=Sum(field))["sum"]
        return Deductible(
            getattr(product, prefix + "_policy", None),
            dedrem_type,
            prev if prev else 0
        )
    return None


# This method is replicating the step 2 of the stored procedure mostly as-is. It will be refactored in several steps.
# The overall process is:
# - Check each product associated with the claim, compute ceilings and maxes
# - Go through each item and deduce
# - Go through each service and deduce
def process_dedrem(claim, audit_user_id=-1, is_process=False):
    logger.debug(f"processing dedrem for claim {claim.uuid}")
    target_date = __get_claim_target_date(claim)
    category = get_claim_category(claim)
    if claim.date_from != target_date:
        hospitalization = True
    else:
        hospitalization = False
    hf_level = claim.health_facility.level

    prev_deductible = None
    prev_remunerated = 0
    prev_remunerated_consult = 0
    prev_remunerated_surgery = 0
    prev_remunerated_hospitalization = 0
    prev_remunerated_delivery = 0
    prev_remunerated_antenatal = 0
    deducted = 0
    remunerated = 0
    remunerated_consultation = 0
    remunerated_surgery = 0
    remunerated_hospitalization = 0
    remunerated_delivery = 0
    remunerated_antenatal = 0
    relative_prices = False

    # TODO: it is not clear in the original code which policy_id was actually used, the latest one apparently...
    policy = None
    ceiling_interpretation = None

    # The original code has a pretty complex query here called product_loop that refers to policies while it is
    # actually looping on ClaimItem and ClaimService.
    items_query = claim.items.filter(
        *filter_validity(validity=target_date, prefix='item__'),
        *filter_validity(),
        *filter_validity(prefix='product__'),
        rejection_reason=0,
    ).values("policy_id", "product_id")
    services_query = claim.services.filter(
        *filter_validity(validity=target_date, prefix='service__'),
        *filter_validity(),
        *filter_validity(prefix='product__'),
        rejection_reason=0,
    ).values("policy_id", "product_id")
    if items_query.count() == 0 and services_query.count() == 0:
        logger.warning(f"claim {claim.uuid} did not have any item or service to valuate.")
    for policy_product in items_query.union(services_query, all=True):
        product = Product.objects.get(*filter_validity(validity=target_date), 
                Q(Q(id=policy_product["product_id"])|Q(legacy_id=policy_product["product_id"])))
        policy_members = InsureePolicy.objects.filter(
            policy_id=policy_product["policy_id"],
            effective_date__isnull=False,
            effective_date__lte=target_date,
            expiry_date__gte=target_date,
            validity_to__isnull=True
        ).count()

        # TODO see declaration of policy_id above
        policy = Policy.objects.get(id=policy_product["policy_id"])
        ceiling_interpretation = product.ceiling_interpretation

        deductible = None
        ceiling = None
        # In previous stored procedure, some commented code fetched the amounts from sum(RemDelivery) from
        # tblClaimDedRem where policy_id, insuree_id & claim_id <> this one
        remunerated_consultation = 0
        if product.max_amount_consultation:
            prev_remunerated_consult = 0
        remunerated_surgery = 0
        if product.max_amount_surgery:
            prev_remunerated_surgery = 0
        remunerated_hospitalization = 0
        if product.max_amount_hospitalization:
            if hospitalization:
                prev_remunerated_hospitalization = 0
        remunerated_delivery = 0
        if product.max_amount_delivery:
            prev_remunerated_delivery = 0
        remunerated_antenatal = 0
        if product.max_amount_antenatal:
            prev_remunerated_antenatal = 0

        ded_g = _get_dedrem("ded", "G", "ded_g", product, claim, policy_product["policy_id"])
        if ded_g:
            deductible = ded_g
            prev_deductible = deductible.prev

        rem_g = _get_dedrem("max", "G", "rem_g", product, claim, policy_product["policy_id"])
        if rem_g:
            ceiling = rem_g
            prev_remunerated = rem_g.prev
        if product.max_policy:
            if policy_members > product.threshold:  # Threshold is NOT NULL
                if product.max_policy_extra_member:
                    ceiling = Deductible(
                        product.max_policy + (policy_members - product.threshold) * product.max_policy_extra_member,
                        ceiling.type,
                        ceiling.prev)
                if product.max_ceiling_policy and ceiling.amount > product.max_ceiling_policy:
                    ceiling = Deductible(
                        product.max_ceiling_policy,
                        ceiling.type,
                        ceiling.prev)
            else:
                ceiling = Deductible(
                    product.max_policy,
                    ceiling.type,
                    ceiling.prev)

        # Then check IP deductibles
        if not deductible:
            if (product.ceiling_interpretation == 'I' and hospitalization) or \
                    (product.ceiling_interpretation == 'H' and hf_level == 'H'):
                # Hospital IP
                ded_ip = _get_dedrem("ded_ip", "I", "ded_ip", product, claim, policy_product["policy_id"])
                if ded_ip:
                    deductible = ded_ip
                    prev_deductible = ded_ip.prev
            else:
                # Hospital OP
                ded_op = _get_dedrem("ded_op", "O", "ded_op", product, claim, policy_product["policy_id"])
                if ded_op:
                    deductible = ded_op
                    prev_deductible = ded_op.prev

        if not ceiling:
            if (product.ceiling_interpretation == 'I' and hospitalization) or \
                    (product.ceiling_interpretation == 'H' and hf_level == 'H'):
                max_ip = _get_dedrem("max_ip", "I", "rem_ip", product, claim, policy_product["policy_id"])
                if max_ip:
                    ceiling = max_ip
                    prev_remunerated = max_ip.prev
                if product.max_ip_policy:
                    if policy_members > product.threshold:  # Threshold is NOT NULL
                        if product.max_policy_extra_member_ip:
                            ceiling = Deductible(
                                product.max_ip_policy + (
                                        policy_members - product.threshold) * product.max_policy_extra_member_ip,
                                ceiling.type,
                                ceiling.prev
                            )
                        if product.max_ceiling_policy_ip and ceiling.amount > product.max_ceiling_policy_ip:
                            ceiling = Deductible(
                                product.max_ceiling_policy_ip,
                                ceiling.type,
                                ceiling.prev
                            )
                    else:
                        ceiling = Deductible(
                            product.max_ip_policy,
                            ceiling.type,
                            ceiling.prev
                        )
            else:
                max_op = _get_dedrem("max_op", "O", "rem_op", product, claim, policy_product["policy_id"])
                if max_op:
                    ceiling = max_op
                    prev_remunerated = max_op.prev
                if product.max_op_policy:
                    if product.threshold and policy_members > product.threshold:
                        if product.max_policy_extra_member_op:
                            ceiling = Deductible(
                                product.max_op_policy + (
                                        policy_members - product.threshold) * product.max_policy_extra_member_op,
                                ceiling.type,
                                ceiling.prev
                            )
                        if product.max_ceiling_policy_op and ceiling.amount > product.max_ceiling_policy_op:
                            ceiling = Deductible(
                                product.max_ceiling_policy_op,
                                ceiling.type,
                                ceiling.prev
                            )
                    else:
                        ceiling = Deductible(
                            product.max_op_policy,
                            ceiling.type,
                            ceiling.prev
                        )

        # Loop through items
        deducted = 0
        remunerated = 0
        for claim_detail in itertools.chain(
                claim.items
                        .filter(validity_to__isnull=True,
                                status=ClaimItem.STATUS_PASSED),
                claim.services
                        .filter(validity_to__isnull=True,
                                status=ClaimService.STATUS_PASSED)):
            detail_is_item = isinstance(claim_detail, ClaimItem)
            itemsvc_quantity = claim_detail.qty_approved \
                if claim_detail.qty_approved is not None else claim_detail.qty_provided
            set_price_deducted = 0
            exceed_ceiling_amount = 0
            exceed_ceiling_amount_category = 0
            # TODO make sure that this does not return more than one row ?
            itemsvc_pricelist_detail_qs = (ItemsPricelistDetail if detail_is_item else ServicesPricelistDetail).objects \
                .filter(itemsvcs_pricelist=claim.health_facility.items_pricelist
            if detail_is_item else claim.health_facility.services_pricelist,
                        itemsvc=claim_detail.itemsvc,
                        itemsvcs_pricelist__validity_to__isnull=True,
                        )
            itemsvc_pricelist_detail = get_queryset_valid_at_date(itemsvc_pricelist_detail_qs, target_date).first()
            product_itemsvc = None

            if detail_is_item:
                product_itemsvc = ProductItem.objects.filter(
                    product_id=claim_detail.product_id,
                    item_id=claim_detail.item_id,
                    validity_to__isnull=True
                ).first()
                if product_itemsvc is None:
                    raise ValueError("Product Item not found")
            else:
                product_itemsvc = ProductService.objects.filter(
                    product=claim_detail.product,
                    service_id=claim_detail.service_id,
                    validity_to__isnull=True
                ).first()
                if product_itemsvc is None:
                    raise ValueError("Product Service not found")

            pl_price = itemsvc_pricelist_detail.price_overrule if itemsvc_pricelist_detail.price_overrule \
                else claim_detail.itemsvc.price

            if claim_detail.price_approved is not None:
                set_price_adjusted = claim_detail.price_approved
            if claim_detail.price_origin == ProductItemOrService.ORIGIN_CLAIM:
                set_price_adjusted = claim_detail.price_asked
                if ClaimConfig.native_code_for_services == False:
                    try:
                        if claim_detail.service.packagetype == 'F':
                            service_price = claim_detail.service.price
                            if claim_detail.price_adjusted is not None:
                                logger.debug(f"compare {claim_detail.price_adjusted} and {service_price}")
                                if claim_detail.price_adjusted > service_price:
                                    set_price_adjusted = service_price
                            else:
                                logger.debug(f"compare {claim_detail.price_asked} and {service_price}")
                                if claim_detail.price_asked > service_price:
                                    set_price_adjusted = service_price
                    except:
                        logger.debug("This it an item element")
            else:
                set_price_adjusted = pl_price
                if ClaimConfig.native_code_for_services == False:
                    try:
                        contunue_service_check = True
                        if claim_detail.service.packagetype == 'P':
                            service_services = ServiceService.objects.filter(servicelinkedService=claim_detail.service.id).all()
                            claim_service_services = ClaimServiceService.objects.filter(claim_service=claim_detail.id).all()
                            if len(service_services) == len(claim_service_services):
                                for servservice in service_services:
                                    for claimserviceservice in claim_service_services:
                                        if servservice.service.id == claimserviceservice.service.id:
                                            logger.debug(f"comparing serviceservice qty {servservice.qty_provided}\
                                                and claimserviceservice qty {claimserviceservice.qty_displayed}")
                                            if servservice.qty_provided != claimserviceservice.qty_displayed:
                                                set_price_adjusted = 0
                                                contunue_service_check = False
                                                break
                                    if contunue_service_check == False:
                                        break
                            else:
                                # user misconfiguration !
                                set_price_adjusted = 0
                                contunue_service_check = False
                            logger.debug(f"set_price_adjusted after service check {set_price_adjusted}")
                            if contunue_service_check:
                                contunue_item_check = True
                                service_items = ServiceItem.objects.filter(servicelinkedItem=claim_detail.service.id).all()
                                claim_service_items = ClaimServiceItem.objects.filter(claim_service=claim_detail.id).all()
                                logger.debug(f"service_items: {service_items}")
                                logger.debug(f"claim_service_items: {claim_service_items}")
                                if len(service_items) == len(claim_service_items):
                                    for serviceitem in service_items:
                                        for claimservicesitem in claim_service_items:
                                            if serviceitem.item.id == claimservicesitem.item.id:
                                                logger.debug(f"comparing serviceitem qty {serviceitem.qty_provided}\
                                                 and claimservicesitem qty {claimservicesitem.qty_displayed}")
                                                if serviceitem.qty_provided != claimservicesitem.qty_displayed:
                                                    set_price_adjusted = 0
                                                    contunue_item_check = False
                                                    break
                                        if contunue_item_check == False:
                                            break
                                else:
                                    # user misconfiguration !
                                    set_price_adjusted = 0
                                logger.debug(f"set_price_adjusted after items check {set_price_adjusted}")
                    except:
                        logger.debug("This is a ClaimItem element, not a ClaimService")

            work_value = int(itemsvc_quantity * set_price_adjusted)

            if claim_detail.limitation == ProductItemOrService.LIMIT_FIXED_AMOUNT \
                    and claim_detail.limitation_value \
                    and (itemsvc_quantity * claim_detail.limitation_value) < work_value:
                work_value = itemsvc_quantity * claim_detail.limitation_value

            if deductible and deductible.amount - prev_deductible - deducted > 0:
                if deductible.amount - deductible.prev - deducted >= work_value:
                    set_price_deducted = work_value
                    deducted += work_value
                    # remunerated += 0 # why ?
                    set_price_valuated = 0
                    set_price_remunerated = 0
                else:
                    # partial coverage
                    set_price_deducted = deductible.amount - deductible.prev - deducted
                    work_value -= set_price_deducted
                    deducted += deductible.amount - deductible.prev - deducted

            if claim_detail.limitation == ProductItemOrService.LIMIT_CO_INSURANCE and claim_detail.limitation_value:
                work_value = claim_detail.limitation_value / 100 * work_value

            if category != Service.CATEGORY_VISIT:
                if product.max_amount_surgery and category == Service.CATEGORY_SURGERY:
                    if work_value + prev_remunerated_surgery + remunerated_surgery <= product.max_amount_surgery:
                        remunerated_surgery += work_value
                    else:
                        if prev_remunerated_surgery + remunerated_surgery >= product.max_amount_surgery:
                            exceed_ceiling_amount_category = work_value
                            # remunerated_surgery += 0
                            work_value = 0
                        else:
                            exceed_ceiling_amount_category = work_value + prev_remunerated_surgery \
                                                             + remunerated_surgery - product.max_amount_surgery
                            work_value -= exceed_ceiling_amount_category
                            remunerated_surgery += work_value
                if product.max_amount_delivery and category == Service.CATEGORY_DELIVERY:
                    if work_value + prev_remunerated_delivery + remunerated_delivery <= product.max_amount_delivery:
                        remunerated_delivery += work_value
                    else:
                        if prev_remunerated_delivery + remunerated_delivery >= product.max_amount_delivery:
                            exceed_ceiling_amount_category = work_value
                            # remunerated_delivery += 0
                            work_value = 0
                        else:
                            exceed_ceiling_amount_category = work_value + prev_remunerated_delivery \
                                                             + remunerated_delivery - product.max_amount_delivery
                            work_value -= exceed_ceiling_amount_category
                            remunerated_delivery += work_value
                if product.max_amount_antenatal and category == Service.CATEGORY_ANTENATAL:
                    if work_value + prev_remunerated_antenatal + remunerated_antenatal <= product.max_amount_antenatal:
                        remunerated_antenatal += work_value
                    else:
                        if prev_remunerated_antenatal + remunerated_antenatal >= product.max_amount_antenatal:
                            exceed_ceiling_amount_category = work_value
                            # remunerated_antenatal += 0
                            work_value = 0
                        else:
                            exceed_ceiling_amount_category = work_value + prev_remunerated_antenatal \
                                                             + remunerated_antenatal - product.max_amount_antenatal
                            work_value -= exceed_ceiling_amount_category
                            remunerated_antenatal += work_value
                if product.max_amount_hospitalization and category == Service.CATEGORY_HOSPITALIZATION:
                    if work_value + prev_remunerated_hospitalization + remunerated_hospitalization \
                            <= product.max_amount_hospitalization:
                        remunerated_hospitalization += work_value
                    else:
                        if prev_remunerated_hospitalization + remunerated_hospitalization \
                                >= product.max_amount_hospitalization:
                            exceed_ceiling_amount_category = work_value
                            # remunerated_hospitalization += 0
                            work_value = 0
                        else:
                            exceed_ceiling_amount_category = work_value + prev_remunerated_hospitalization \
                                                             + remunerated_hospitalization \
                                                             - product.max_amount_hospitalization
                            work_value -= exceed_ceiling_amount_category
                            remunerated_hospitalization += work_value
                if product.max_amount_consultation and category == Service.CATEGORY_CONSULTATION:
                    if work_value + prev_remunerated_consult + remunerated_consultation \
                            <= product.max_amount_consultation:
                        remunerated_consultation += work_value
                    else:
                        if prev_remunerated_consult + remunerated_consultation >= product.max_amount_consultation:
                            exceed_ceiling_amount_category = work_value
                            # remunerated_consult += 0
                            work_value = 0
                        else:
                            exceed_ceiling_amount_category = work_value + prev_remunerated_consult \
                                                             + remunerated_consultation - product.max_amount_consultation
                            work_value -= exceed_ceiling_amount_category
                            remunerated_consultation += work_value

            # TODO big rework of this condition is needed. putting the ceiling_exclusion_? into a variable as first step
            if (product_itemsvc is not None and claim.insuree.is_adult
                and (
                        (
                                product.ceiling_interpretation == Product.CEILING_INTERPRETATION_IN_PATIENT and hospitalization == 1)
                        or (
                                product.ceiling_interpretation == Product.CEILING_INTERPRETATION_HOSPITAL and hf_level == "H")
                )
                and product_itemsvc.ceiling_exclusion_adult in ("B", "H")
            ) or (claim.insuree.is_adult
                  and not (
                            (
                                    product.ceiling_interpretation == Product.CEILING_INTERPRETATION_IN_PATIENT and hospitalization == 1)
                            or (
                                    product.ceiling_interpretation == Product.CEILING_INTERPRETATION_HOSPITAL and hf_level == "H")
                    )
                  and product_itemsvc.ceiling_exclusion_adult in ("B", "N")
            ) or (not claim.insuree.is_adult
                  and (
                          (
                                  product.ceiling_interpretation == Product.CEILING_INTERPRETATION_IN_PATIENT and hospitalization == 1)
                          or (
                                  product.ceiling_interpretation == Product.CEILING_INTERPRETATION_HOSPITAL and hf_level == "H")
                  )
                  and product_itemsvc.ceiling_exclusion_child in ("B", "H")
            ) or (not claim.insuree.is_adult
                  and not (
                            (
                                    product.ceiling_interpretation == Product.CEILING_INTERPRETATION_IN_PATIENT and hospitalization == 1)
                            or (
                                    product.ceiling_interpretation == Product.CEILING_INTERPRETATION_HOSPITAL and hf_level == "H")
                    )
                  and product_itemsvc.ceiling_exclusion_child in ("B", "N")
            ):
                # NO CEILING WILL BE AFFECTED
                exceed_ceiling_amount = 0
                # remunerated += 0
                # here in this case we do not add the amount to be added to the
                # ceiling --> so exclude from the actual value to be entered against the insert into tblClaimDedRem
                # in the end of the prod loop
                set_price_valuated = work_value
                set_price_remunerated = work_value
            else:
                if ceiling and ceiling.amount > 0:
                    if ceiling.amount - prev_remunerated - remunerated > 0:
                        if ceiling.amount - prev_remunerated - remunerated >= work_value:
                            exceed_ceiling_amount = 0
                            set_price_valuated = work_value
                            set_price_remunerated = work_value
                            remunerated += work_value
                        else:
                            total = ceiling.amount - prev_remunerated - remunerated
                            exceed_ceiling_amount = work_value - total
                            set_price_valuated = total
                            set_price_remunerated = total
                            remunerated += total
                    else:
                        exceed_ceiling_amount = work_value
                        # remunerated += 0
                        set_price_valuated = 0
                        set_price_remunerated = 0
                else:
                    exceed_ceiling_amount = 0
                    remunerated += work_value
                    set_price_valuated = work_value
                    set_price_remunerated = work_value

            # TODO here was NextItem target. Some "goto nextitem" above might have been replaced with a continue instead
            if is_process:
                if claim_detail.price_origin == ProductItemOrService.ORIGIN_RELATIVE:
                    claim_detail.price_adjusted = set_price_adjusted
                    claim_detail.price_valuated = set_price_valuated
                    claim_detail.deductable_amount = set_price_deducted
                    claim_detail.exceed_ceiling_amount = exceed_ceiling_amount
                    # TODO ExceedCeilingAmountCategory = ExceedCeilingAmountCategory ???
                    relative_prices = True
                else:
                    claim_detail.price_adjusted = set_price_adjusted
                    claim_detail.price_valuated = set_price_valuated
                    claim_detail.deductable_amount = set_price_deducted
                    claim_detail.exceed_ceiling_amount = exceed_ceiling_amount
                    # TODO ExceedCeilingAmountCategory = ExceedCeilingAmountCategory ???
                    claim_detail.remunerated_amount = set_price_remunerated
                    # Don't touch relative_prices
                claim_detail.save()

    # amount is 'locked' from the submit
    # ... so re-creating the ClaimDedRem according to adjusted/valuated price
    ClaimDedRem.objects.filter(claim=claim).delete()

    from core import datetime
    now = datetime.datetime.now()
    claim_ded_rem_to_create = {
        "policy": policy,
        "insuree": claim.insuree,
        "claim": claim,
        "ded_g": deducted,
        "rem_g": remunerated,
        "rem_consult": remunerated_consultation,
        "rem_hospitalization": remunerated_hospitalization,
        "rem_delivery": remunerated_delivery,
        "rem_antenatal": remunerated_antenatal,
        "rem_surgery": remunerated_surgery,
        "audit_user_id": audit_user_id,
        "validity_from": now
    }
    if (ceiling_interpretation == "I" and hospitalization == 1) or (ceiling_interpretation == "H" and hf_level == "H"):
        claim_ded_rem_to_create["ded_ip"] = deducted
        claim_ded_rem_to_create["rem_ip"] = remunerated
    else:
        claim_ded_rem_to_create["ded_op"] = deducted
        claim_ded_rem_to_create["rem_op"] = remunerated

    ClaimDedRem.objects.create(**claim_ded_rem_to_create)

    if is_process:
        if relative_prices:
            claim.status = Claim.STATUS_PROCESSED
        else:
            claim.status = Claim.STATUS_VALUATED
        claim.audit_user_id_process = audit_user_id
        claim.process_stamp = now
        claim.date_processed = now
        if claim.feedback_status == Claim.FEEDBACK_SELECTED:
            claim.feedback_status = Claim.FEEDBACK_BYPASSED
        if claim.review_status == Claim.REVIEW_SELECTED:
            claim.review_status = Claim.REVIEW_BYPASSED

        claim.save()

    return []  # process_dedrem will never put the claim in error status (beside technical error and until it changes)



# Generated by Django 3.2.16 on 2023-07-05 14:01

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('claim', '0019_auto_20230615_1055'),
    ]

    operations = [
        migrations.AlterField(
            model_name='claim',
            name='code',
            field=models.CharField(db_column='ClaimCode', max_length=50),
        ),
    ]



from django.db import connection
from tools.utils import dictfetchall

import logging

logger = logging.getLogger(__name__)

template = """
{
    "docElements": [
        {
            "elementType": "text",
            "id": 3,
            "containerId": "0_header",
            "x": 0,
            "y": 20,
            "width": 575,
            "height": 40,
            "content": "Percentage of referrals in claims",
            "richText": false,
            "richTextContent": null,
            "richTextHtml": "",
            "eval": false,
            "styleId": "",
            "bold": true,
            "italic": false,
            "underline": false,
            "strikethrough": false,
            "horizontalAlignment": "center",
            "verticalAlignment": "middle",
            "textColor": "#006374",
            "backgroundColor": "",
            "font": "helvetica",
            "fontSize": "24",
            "lineSpacing": 1,
            "borderColor": "#000000",
            "borderWidth": 1,
            "borderAll": false,
            "borderLeft": false,
            "borderTop": false,
            "borderRight": false,
            "borderBottom": false,
            "paddingLeft": 2,
            "paddingTop": 2,
            "paddingRight": 2,
            "paddingBottom": 2,
            "printIf": "",
            "removeEmptyElement": false,
            "alwaysPrintOnSamePage": true,
            "pattern": "",
            "link": "",
            "cs_condition": "",
            "cs_styleId": "",
            "cs_bold": false,
            "cs_italic": false,
            "cs_underline": false,
            "cs_strikethrough": false,
            "cs_horizontalAlignment": "left",
            "cs_verticalAlignment": "top",
            "cs_textColor": "#000000",
            "cs_backgroundColor": "",
            "cs_font": "helvetica",
            "cs_fontSize": 12,
            "cs_lineSpacing": 1,
            "cs_borderColor": "#000000",
            "cs_borderWidth": "1",
            "cs_borderAll": false,
            "cs_borderLeft": false,
            "cs_borderTop": false,
            "cs_borderRight": false,
            "cs_borderBottom": false,
            "cs_paddingLeft": 2,
            "cs_paddingTop": 2,
            "cs_paddingRight": 2,
            "cs_paddingBottom": 2,
            "spreadsheet_hide": false,
            "spreadsheet_column": "",
            "spreadsheet_colspan": "",
            "spreadsheet_addEmptyRow": false,
            "spreadsheet_textWrap": false
        },
        {
            "elementType": "line",
            "id": 158,
            "containerId": "0_content",
            "x": 0,
            "y": 0,
            "width": 575,
            "height": 1,
            "color": "#000000",
            "printIf": ""
        },
        {
            "elementType": "text",
            "id": 193,
            "containerId": "0_content",
            "x": 0,
            "y": 10,
            "width": 575,
            "height": 20,
            "content": "List of health facilities, along with total number of claims, outpatient referrals total and inpatient referrals totals",
            "richText": false,
            "richTextContent": null,
            "richTextHtml": "",
            "eval": false,
            "styleId": "",
            "bold": true,
            "italic": false,
            "underline": true,
            "strikethrough": false,
            "horizontalAlignment": "center",
            "verticalAlignment": "top",
            "textColor": "#000000",
            "backgroundColor": "",
            "font": "helvetica",
            "fontSize": "10",
            "lineSpacing": 1,
            "borderColor": "#000000",
            "borderWidth": 1,
            "borderAll": false,
            "borderLeft": false,
            "borderTop": false,
            "borderRight": false,
            "borderBottom": false,
            "paddingLeft": 2,
            "paddingTop": 2,
            "paddingRight": 2,
            "paddingBottom": 2,
            "printIf": "",
            "removeEmptyElement": false,
            "alwaysPrintOnSamePage": true,
            "pattern": "",
            "link": "",
            "cs_condition": "",
            "cs_styleId": "",
            "cs_bold": false,
            "cs_italic": false,
            "cs_underline": false,
            "cs_strikethrough": false,
            "cs_horizontalAlignment": "left",
            "cs_verticalAlignment": "top",
            "cs_textColor": "#000000",
            "cs_backgroundColor": "",
            "cs_font": "helvetica",
            "cs_fontSize": 12,
            "cs_lineSpacing": 1,
            "cs_borderColor": "#000000",
            "cs_borderWidth": "1",
            "cs_borderAll": false,
            "cs_borderLeft": false,
            "cs_borderTop": false,
            "cs_borderRight": false,
            "cs_borderBottom": false,
            "cs_paddingLeft": 2,
            "cs_paddingTop": 2,
            "cs_paddingRight": 2,
            "cs_paddingBottom": 2,
            "spreadsheet_hide": false,
            "spreadsheet_column": "",
            "spreadsheet_colspan": "",
            "spreadsheet_addEmptyRow": false,
            "spreadsheet_textWrap": false
        },
        {
            "elementType": "table",
            "id": 272,
            "containerId": "0_content",
            "width": 574,
            "x": 0,
            "y": 40,
            "dataSource": "${data}",
            "columns": 4,
            "header": true,
            "contentRows": "1",
            "footer": false,
            "border": "grid",
            "borderColor": "#000000",
            "borderWidth": "1",
            "printIf": "",
            "removeEmptyElement": false,
            "spreadsheet_hide": false,
            "spreadsheet_column": "",
            "spreadsheet_addEmptyRow": false,
            "headerData": {
                "elementType": "none",
                "id": 273,
                "height": 20,
                "backgroundColor": "",
                "repeatHeader": false,
                "columnData": [
                    {
                        "elementType": "table_text",
                        "id": 274,
                        "width": 304,
                        "content": "Health Facility",
                        "eval": false,
                        "colspan": "",
                        "styleId": "",
                        "bold": true,
                        "italic": false,
                        "underline": false,
                        "strikethrough": false,
                        "horizontalAlignment": "center",
                        "verticalAlignment": "middle",
                        "textColor": "#000000",
                        "backgroundColor": "",
                        "font": "helvetica",
                        "fontSize": 12,
                        "lineSpacing": 1,
                        "paddingLeft": 2,
                        "paddingTop": 2,
                        "paddingRight": 2,
                        "paddingBottom": 2,
                        "pattern": "",
                        "link": "",
                        "cs_condition": "",
                        "cs_styleId": "",
                        "cs_bold": false,
                        "cs_italic": false,
                        "cs_underline": false,
                        "cs_strikethrough": false,
                        "cs_horizontalAlignment": "left",
                        "cs_verticalAlignment": "top",
                        "cs_textColor": "#000000",
                        "cs_backgroundColor": "",
                        "cs_font": "helvetica",
                        "cs_fontSize": 12,
                        "cs_lineSpacing": 1,
                        "cs_paddingLeft": 2,
                        "cs_paddingTop": 2,
                        "cs_paddingRight": 2,
                        "cs_paddingBottom": 2,
                        "spreadsheet_textWrap": false,
                        "printIf": "",
                        "growWeight": 0,
                        "borderWidth": 1
                    },
                    {
                        "elementType": "table_text",
                        "id": 275,
                        "width": 90,
                        "content": "Total Claims",
                        "eval": false,
                        "colspan": "",
                        "styleId": "",
                        "bold": true,
                        "italic": false,
                        "underline": false,
                        "strikethrough": false,
                        "horizontalAlignment": "center",
                        "verticalAlignment": "middle",
                        "textColor": "#000000",
                        "backgroundColor": "",
                        "font": "helvetica",
                        "fontSize": 12,
                        "lineSpacing": 1,
                        "paddingLeft": 2,
                        "paddingTop": 2,
                        "paddingRight": 2,
                        "paddingBottom": 2,
                        "pattern": "",
                        "link": "",
                        "cs_condition": "",
                        "cs_styleId": "",
                        "cs_bold": false,
                        "cs_italic": false,
                        "cs_underline": false,
                        "cs_strikethrough": false,
                        "cs_horizontalAlignment": "left",
                        "cs_verticalAlignment": "top",
                        "cs_textColor": "#000000",
                        "cs_backgroundColor": "",
                        "cs_font": "helvetica",
                        "cs_fontSize": 12,
                        "cs_lineSpacing": 1,
                        "cs_paddingLeft": 2,
                        "cs_paddingTop": 2,
                        "cs_paddingRight": 2,
                        "cs_paddingBottom": 2,
                        "spreadsheet_textWrap": false,
                        "printIf": "",
                        "growWeight": 0,
                        "borderWidth": 1
                    },
                    {
                        "elementType": "table_text",
                        "id": 282,
                        "width": 90,
                        "content": "Referrals (OP)",
                        "eval": false,
                        "colspan": "",
                        "styleId": "",
                        "bold": true,
                        "italic": false,
                        "underline": false,
                        "strikethrough": false,
                        "horizontalAlignment": "center",
                        "verticalAlignment": "middle",
                        "textColor": "#000000",
                        "backgroundColor": "",
                        "font": "helvetica",
                        "fontSize": 12,
                        "lineSpacing": 1,
                        "paddingLeft": 2,
                        "paddingTop": 2,
                        "paddingRight": 2,
                        "paddingBottom": 2,
                        "pattern": "",
                        "link": "",
                        "cs_condition": "",
                        "cs_styleId": "",
                        "cs_bold": false,
                        "cs_italic": false,
                        "cs_underline": false,
                        "cs_strikethrough": false,
                        "cs_horizontalAlignment": "left",
                        "cs_verticalAlignment": "top",
                        "cs_textColor": "#000000",
                        "cs_backgroundColor": "",
                        "cs_font": "helvetica",
                        "cs_fontSize": 12,
                        "cs_lineSpacing": 1,
                        "cs_paddingLeft": 2,
                        "cs_paddingTop": 2,
                        "cs_paddingRight": 2,
                        "cs_paddingBottom": 2,
                        "spreadsheet_textWrap": false,
                        "printIf": "",
                        "growWeight": 0,
                        "borderWidth": 1
                    },
                    {
                        "elementType": "table_text",
                        "id": 285,
                        "width": 90,
                        "content": "Referrals (IP)",
                        "eval": false,
                        "colspan": "",
                        "styleId": "",
                        "bold": true,
                        "italic": false,
                        "underline": false,
                        "strikethrough": false,
                        "horizontalAlignment": "center",
                        "verticalAlignment": "middle",
                        "textColor": "#000000",
                        "backgroundColor": "",
                        "font": "helvetica",
                        "fontSize": 12,
                        "lineSpacing": 1,
                        "paddingLeft": 2,
                        "paddingTop": 2,
                        "paddingRight": 2,
                        "paddingBottom": 2,
                        "pattern": "",
                        "link": "",
                        "cs_condition": "",
                        "cs_styleId": "",
                        "cs_bold": false,
                        "cs_italic": false,
                        "cs_underline": false,
                        "cs_strikethrough": false,
                        "cs_horizontalAlignment": "left",
                        "cs_verticalAlignment": "top",
                        "cs_textColor": "#000000",
                        "cs_backgroundColor": "",
                        "cs_font": "helvetica",
                        "cs_fontSize": 12,
                        "cs_lineSpacing": 1,
                        "cs_paddingLeft": 2,
                        "cs_paddingTop": 2,
                        "cs_paddingRight": 2,
                        "cs_paddingBottom": 2,
                        "spreadsheet_textWrap": false,
                        "printIf": "",
                        "growWeight": 0,
                        "borderWidth": 1
                    }
                ]
            },
            "contentDataRows": [
                {
                    "elementType": "none",
                    "id": 276,
                    "height": 20,
                    "backgroundColor": "",
                    "alternateBackgroundColor": "",
                    "groupExpression": "",
                    "printIf": "",
                    "alwaysPrintOnSamePage": true,
                    "pageBreak": false,
                    "repeatGroupHeader": false,
                    "columnData": [
                        {
                            "elementType": "table_text",
                            "id": 277,
                            "width": 304,
                            "content": "${hf}",
                            "eval": false,
                            "colspan": "",
                            "styleId": "",
                            "bold": false,
                            "italic": false,
                            "underline": false,
                            "strikethrough": false,
                            "horizontalAlignment": "left",
                            "verticalAlignment": "middle",
                            "textColor": "#000000",
                            "backgroundColor": "",
                            "font": "helvetica",
                            "fontSize": 12,
                            "lineSpacing": 1,
                            "paddingLeft": 2,
                            "paddingTop": 2,
                            "paddingRight": 2,
                            "paddingBottom": 2,
                            "pattern": "",
                            "link": "",
                            "cs_condition": "",
                            "cs_styleId": "",
                            "cs_bold": false,
                            "cs_italic": false,
                            "cs_underline": false,
                            "cs_strikethrough": false,
                            "cs_horizontalAlignment": "left",
                            "cs_verticalAlignment": "top",
                            "cs_textColor": "#000000",
                            "cs_backgroundColor": "",
                            "cs_font": "helvetica",
                            "cs_fontSize": 12,
                            "cs_lineSpacing": 1,
                            "cs_paddingLeft": 2,
                            "cs_paddingTop": 2,
                            "cs_paddingRight": 2,
                            "cs_paddingBottom": 2,
                            "spreadsheet_textWrap": false,
                            "borderWidth": 1,
                            "growWeight": 0
                        },
                        {
                            "elementType": "table_text",
                            "id": 278,
                            "width": 90,
                            "content": "${totalclaims}",
                            "eval": false,
                            "colspan": "",
                            "styleId": "",
                            "bold": false,
                            "italic": false,
                            "underline": false,
                            "strikethrough": false,
                            "horizontalAlignment": "center",
                            "verticalAlignment": "middle",
                            "textColor": "#000000",
                            "backgroundColor": "",
                            "font": "helvetica",
                            "fontSize": 12,
                            "lineSpacing": 1,
                            "paddingLeft": 2,
                            "paddingTop": 2,
                            "paddingRight": 2,
                            "paddingBottom": 2,
                            "pattern": "",
                            "link": "",
                            "cs_condition": "",
                            "cs_styleId": "",
                            "cs_bold": false,
                            "cs_italic": false,
                            "cs_underline": false,
                            "cs_strikethrough": false,
                            "cs_horizontalAlignment": "left",
                            "cs_verticalAlignment": "top",
                            "cs_textColor": "#000000",
                            "cs_backgroundColor": "",
                            "cs_font": "helvetica",
                            "cs_fontSize": 12,
                            "cs_lineSpacing": 1,
                            "cs_paddingLeft": 2,
                            "cs_paddingTop": 2,
                            "cs_paddingRight": 2,
                            "cs_paddingBottom": 2,
                            "spreadsheet_textWrap": false,
                            "borderWidth": 1,
                            "growWeight": 0
                        },
                        {
                            "elementType": "table_text",
                            "id": 283,
                            "width": 90,
                            "content": "${totalop}",
                            "eval": false,
                            "colspan": "",
                            "styleId": "",
                            "bold": false,
                            "italic": false,
                            "underline": false,
                            "strikethrough": false,
                            "horizontalAlignment": "center",
                            "verticalAlignment": "middle",
                            "textColor": "#000000",
                            "backgroundColor": "",
                            "font": "helvetica",
                            "fontSize": 12,
                            "lineSpacing": 1,
                            "paddingLeft": 2,
                            "paddingTop": 2,
                            "paddingRight": 2,
                            "paddingBottom": 2,
                            "pattern": "",
                            "link": "",
                            "cs_condition": "",
                            "cs_styleId": "",
                            "cs_bold": false,
                            "cs_italic": false,
                            "cs_underline": false,
                            "cs_strikethrough": false,
                            "cs_horizontalAlignment": "left",
                            "cs_verticalAlignment": "top",
                            "cs_textColor": "#000000",
                            "cs_backgroundColor": "",
                            "cs_font": "helvetica",
                            "cs_fontSize": 12,
                            "cs_lineSpacing": 1,
                            "cs_paddingLeft": 2,
                            "cs_paddingTop": 2,
                            "cs_paddingRight": 2,
                            "cs_paddingBottom": 2,
                            "spreadsheet_textWrap": false,
                            "borderWidth": 1,
                            "growWeight": 0
                        },
                        {
                            "elementType": "table_text",
                            "id": 286,
                            "width": 90,
                            "content": "${totalip}",
                            "eval": false,
                            "colspan": "",
                            "styleId": "",
                            "bold": false,
                            "italic": false,
                            "underline": false,
                            "strikethrough": false,
                            "horizontalAlignment": "center",
                            "verticalAlignment": "middle",
                            "textColor": "#000000",
                            "backgroundColor": "",
                            "font": "helvetica",
                            "fontSize": 12,
                            "lineSpacing": 1,
                            "paddingLeft": 2,
                            "paddingTop": 2,
                            "paddingRight": 2,
                            "paddingBottom": 2,
                            "pattern": "",
                            "link": "",
                            "cs_condition": "",
                            "cs_styleId": "",
                            "cs_bold": false,
                            "cs_italic": false,
                            "cs_underline": false,
                            "cs_strikethrough": false,
                            "cs_horizontalAlignment": "left",
                            "cs_verticalAlignment": "top",
                            "cs_textColor": "#000000",
                            "cs_backgroundColor": "",
                            "cs_font": "helvetica",
                            "cs_fontSize": 12,
                            "cs_lineSpacing": 1,
                            "cs_paddingLeft": 2,
                            "cs_paddingTop": 2,
                            "cs_paddingRight": 2,
                            "cs_paddingBottom": 2,
                            "spreadsheet_textWrap": false,
                            "borderWidth": 1,
                            "growWeight": 0
                        }
                    ]
                }
            ],
            "footerData": {
                "elementType": "none",
                "id": 279,
                "height": 20,
                "backgroundColor": "",
                "columnData": [
                    {
                        "elementType": "table_text",
                        "id": 280,
                        "width": 304,
                        "content": "",
                        "eval": false,
                        "colspan": "",
                        "styleId": "",
                        "bold": false,
                        "italic": false,
                        "underline": false,
                        "strikethrough": false,
                        "horizontalAlignment": "left",
                        "verticalAlignment": "top",
                        "textColor": "#000000",
                        "backgroundColor": "",
                        "font": "helvetica",
                        "fontSize": 12,
                        "lineSpacing": 1,
                        "paddingLeft": 2,
                        "paddingTop": 2,
                        "paddingRight": 2,
                        "paddingBottom": 2,
                        "pattern": "",
                        "link": "",
                        "cs_condition": "",
                        "cs_styleId": "",
                        "cs_bold": false,
                        "cs_italic": false,
                        "cs_underline": false,
                        "cs_strikethrough": false,
                        "cs_horizontalAlignment": "left",
                        "cs_verticalAlignment": "top",
                        "cs_textColor": "#000000",
                        "cs_backgroundColor": "",
                        "cs_font": "helvetica",
                        "cs_fontSize": 12,
                        "cs_lineSpacing": 1,
                        "cs_paddingLeft": 2,
                        "cs_paddingTop": 2,
                        "cs_paddingRight": 2,
                        "cs_paddingBottom": 2,
                        "spreadsheet_textWrap": false,
                        "borderWidth": 1,
                        "growWeight": 0
                    },
                    {
                        "elementType": "table_text",
                        "id": 281,
                        "width": 90,
                        "content": "",
                        "eval": false,
                        "colspan": "",
                        "styleId": "",
                        "bold": false,
                        "italic": false,
                        "underline": false,
                        "strikethrough": false,
                        "horizontalAlignment": "left",
                        "verticalAlignment": "top",
                        "textColor": "#000000",
                        "backgroundColor": "",
                        "font": "helvetica",
                        "fontSize": 12,
                        "lineSpacing": 1,
                        "paddingLeft": 2,
                        "paddingTop": 2,
                        "paddingRight": 2,
                        "paddingBottom": 2,
                        "pattern": "",
                        "link": "",
                        "cs_condition": "",
                        "cs_styleId": "",
                        "cs_bold": false,
                        "cs_italic": false,
                        "cs_underline": false,
                        "cs_strikethrough": false,
                        "cs_horizontalAlignment": "left",
                        "cs_verticalAlignment": "top",
                        "cs_textColor": "#000000",
                        "cs_backgroundColor": "",
                        "cs_font": "helvetica",
                        "cs_fontSize": 12,
                        "cs_lineSpacing": 1,
                        "cs_paddingLeft": 2,
                        "cs_paddingTop": 2,
                        "cs_paddingRight": 2,
                        "cs_paddingBottom": 2,
                        "spreadsheet_textWrap": false,
                        "borderWidth": 1,
                        "growWeight": 0
                    },
                    {
                        "elementType": "table_text",
                        "id": 284,
                        "width": 90,
                        "content": "",
                        "eval": false,
                        "colspan": "",
                        "styleId": "",
                        "bold": false,
                        "italic": false,
                        "underline": false,
                        "strikethrough": false,
                        "horizontalAlignment": "left",
                        "verticalAlignment": "top",
                        "textColor": "#000000",
                        "backgroundColor": "",
                        "font": "helvetica",
                        "fontSize": 12,
                        "lineSpacing": 1,
                        "paddingLeft": 2,
                        "paddingTop": 2,
                        "paddingRight": 2,
                        "paddingBottom": 2,
                        "pattern": "",
                        "link": "",
                        "cs_condition": "",
                        "cs_styleId": "",
                        "cs_bold": false,
                        "cs_italic": false,
                        "cs_underline": false,
                        "cs_strikethrough": false,
                        "cs_horizontalAlignment": "left",
                        "cs_verticalAlignment": "top",
                        "cs_textColor": "#000000",
                        "cs_backgroundColor": "",
                        "cs_font": "helvetica",
                        "cs_fontSize": 12,
                        "cs_lineSpacing": 1,
                        "cs_paddingLeft": 2,
                        "cs_paddingTop": 2,
                        "cs_paddingRight": 2,
                        "cs_paddingBottom": 2,
                        "spreadsheet_textWrap": false,
                        "borderWidth": 1,
                        "growWeight": 0
                    },
                    {
                        "elementType": "table_text",
                        "id": 287,
                        "width": 90,
                        "content": "",
                        "eval": false,
                        "colspan": "",
                        "styleId": "",
                        "bold": false,
                        "italic": false,
                        "underline": false,
                        "strikethrough": false,
                        "horizontalAlignment": "left",
                        "verticalAlignment": "top",
                        "textColor": "#000000",
                        "backgroundColor": "",
                        "font": "helvetica",
                        "fontSize": 12,
                        "lineSpacing": 1,
                        "paddingLeft": 2,
                        "paddingTop": 2,
                        "paddingRight": 2,
                        "paddingBottom": 2,
                        "pattern": "",
                        "link": "",
                        "cs_condition": "",
                        "cs_styleId": "",
                        "cs_bold": false,
                        "cs_italic": false,
                        "cs_underline": false,
                        "cs_strikethrough": false,
                        "cs_horizontalAlignment": "left",
                        "cs_verticalAlignment": "top",
                        "cs_textColor": "#000000",
                        "cs_backgroundColor": "",
                        "cs_font": "helvetica",
                        "cs_fontSize": 12,
                        "cs_lineSpacing": 1,
                        "cs_paddingLeft": 2,
                        "cs_paddingTop": 2,
                        "cs_paddingRight": 2,
                        "cs_paddingBottom": 2,
                        "spreadsheet_textWrap": false,
                        "borderWidth": 1,
                        "growWeight": 0
                    }
                ]
            }
        },
        {
            "elementType": "text",
            "id": 7,
            "containerId": "0_footer",
            "x": 320,
            "y": 0,
            "width": 255,
            "height": 30,
            "content": "Page ${page_number} / ${page_count}",
            "richText": false,
            "richTextContent": null,
            "richTextHtml": "",
            "eval": false,
            "styleId": "",
            "bold": false,
            "italic": false,
            "underline": false,
            "strikethrough": false,
            "horizontalAlignment": "right",
            "verticalAlignment": "middle",
            "textColor": "#666666",
            "backgroundColor": "",
            "font": "helvetica",
            "fontSize": 12,
            "lineSpacing": 1,
            "borderColor": "#000000",
            "borderWidth": 1,
            "borderAll": false,
            "borderLeft": false,
            "borderTop": false,
            "borderRight": false,
            "borderBottom": false,
            "paddingLeft": 2,
            "paddingTop": 2,
            "paddingRight": 2,
            "paddingBottom": 2,
            "printIf": "",
            "removeEmptyElement": false,
            "alwaysPrintOnSamePage": true,
            "pattern": "",
            "link": "",
            "cs_condition": "",
            "cs_styleId": "",
            "cs_bold": false,
            "cs_italic": false,
            "cs_underline": false,
            "cs_strikethrough": false,
            "cs_horizontalAlignment": "left",
            "cs_verticalAlignment": "top",
            "cs_textColor": "#000000",
            "cs_backgroundColor": "",
            "cs_font": "helvetica",
            "cs_fontSize": 12,
            "cs_lineSpacing": 1,
            "cs_borderColor": "#000000",
            "cs_borderWidth": "1",
            "cs_borderAll": false,
            "cs_borderLeft": false,
            "cs_borderTop": false,
            "cs_borderRight": false,
            "cs_borderBottom": false,
            "cs_paddingLeft": 2,
            "cs_paddingTop": 2,
            "cs_paddingRight": 2,
            "cs_paddingBottom": 2,
            "spreadsheet_hide": false,
            "spreadsheet_column": "",
            "spreadsheet_colspan": "",
            "spreadsheet_addEmptyRow": false,
            "spreadsheet_textWrap": false
        },
        {
            "elementType": "text",
            "id": 8,
            "containerId": "0_footer",
            "x": 0,
            "y": 0,
            "width": 290,
            "height": 30,
            "content": "Created on ${current_date}",
            "richText": false,
            "richTextContent": null,
            "richTextHtml": "",
            "eval": false,
            "styleId": "",
            "bold": false,
            "italic": false,
            "underline": false,
            "strikethrough": false,
            "horizontalAlignment": "left",
            "verticalAlignment": "middle",
            "textColor": "#666666",
            "backgroundColor": "",
            "font": "helvetica",
            "fontSize": 12,
            "lineSpacing": 1,
            "borderColor": "#000000",
            "borderWidth": 1,
            "borderAll": false,
            "borderLeft": false,
            "borderTop": false,
            "borderRight": false,
            "borderBottom": false,
            "paddingLeft": 2,
            "paddingTop": 2,
            "paddingRight": 2,
            "paddingBottom": 2,
            "printIf": "",
            "removeEmptyElement": false,
            "alwaysPrintOnSamePage": true,
            "pattern": "",
            "link": "",
            "cs_condition": "",
            "cs_styleId": "",
            "cs_bold": false,
            "cs_italic": false,
            "cs_underline": false,
            "cs_strikethrough": false,
            "cs_horizontalAlignment": "left",
            "cs_verticalAlignment": "top",
            "cs_textColor": "#000000",
            "cs_backgroundColor": "",
            "cs_font": "helvetica",
            "cs_fontSize": 12,
            "cs_lineSpacing": 1,
            "cs_borderColor": "#000000",
            "cs_borderWidth": "1",
            "cs_borderAll": false,
            "cs_borderLeft": false,
            "cs_borderTop": false,
            "cs_borderRight": false,
            "cs_borderBottom": false,
            "cs_paddingLeft": 2,
            "cs_paddingTop": 2,
            "cs_paddingRight": 2,
            "cs_paddingBottom": 2,
            "spreadsheet_hide": false,
            "spreadsheet_column": "",
            "spreadsheet_colspan": "",
            "spreadsheet_addEmptyRow": false,
            "spreadsheet_textWrap": false
        }
    ],
    "parameters": [
        {
            "id": 1,
            "name": "page_count",
            "type": "number",
            "arrayItemType": "string",
            "eval": false,
            "nullable": false,
            "pattern": "",
            "expression": "",
            "showOnlyNameType": true,
            "testData": ""
        },
        {
            "id": 2,
            "name": "page_number",
            "type": "number",
            "arrayItemType": "string",
            "eval": false,
            "nullable": false,
            "pattern": "",
            "expression": "",
            "showOnlyNameType": true,
            "testData": ""
        },
        {
            "id": 9,
            "name": "current_date",
            "type": "date",
            "arrayItemType": "string",
            "eval": false,
            "nullable": false,
            "pattern": "d/M/yyyy H:mm",
            "expression": "",
            "showOnlyNameType": false,
            "testData": ""
        },
        {
            "id": 233,
            "name": "data",
            "type": "array",
            "arrayItemType": "string",
            "eval": false,
            "nullable": false,
            "pattern": "",
            "expression": "",
            "showOnlyNameType": false,
            "testData": "",
            "children": [
                {
                    "id": 234,
                    "name": "row_number",
                    "type": "number",
                    "arrayItemType": "string",
                    "eval": false,
                    "nullable": false,
                    "pattern": "",
                    "expression": "",
                    "showOnlyNameType": true,
                    "testData": ""
                },
                {
                    "id": 235,
                    "name": "hf",
                    "type": "string",
                    "arrayItemType": "string",
                    "eval": false,
                    "nullable": false,
                    "pattern": "",
                    "expression": "",
                    "showOnlyNameType": false,
                    "testData": ""
                },
                {
                    "id": 236,
                    "name": "totalclaims",
                    "type": "number",
                    "arrayItemType": "string",
                    "eval": false,
                    "nullable": true,
                    "pattern": "#,##0",
                    "expression": "",
                    "showOnlyNameType": false,
                    "testData": ""
                },
                {
                    "id": 262,
                    "name": "totalop",
                    "type": "number",
                    "arrayItemType": "string",
                    "eval": false,
                    "nullable": true,
                    "pattern": "#,##0",
                    "expression": "",
                    "showOnlyNameType": false,
                    "testData": ""
                },
                {
                    "id": 271,
                    "name": "totalip",
                    "type": "number",
                    "arrayItemType": "string",
                    "eval": false,
                    "nullable": true,
                    "pattern": "#,##0",
                    "expression": "",
                    "showOnlyNameType": false,
                    "testData": ""
                }
            ]
        }
    ],
    "styles": [
        {
            "id": 33,
            "name": "Table Header",
            "bold": true,
            "italic": false,
            "underline": false,
            "strikethrough": false,
            "horizontalAlignment": "left",
            "verticalAlignment": "middle",
            "textColor": "#000000",
            "backgroundColor": "",
            "font": "helvetica",
            "fontSize": "12",
            "lineSpacing": 1,
            "borderColor": "#000000",
            "borderWidth": "1",
            "borderAll": false,
            "borderLeft": false,
            "borderTop": false,
            "borderRight": false,
            "borderBottom": false,
            "paddingLeft": "2",
            "paddingTop": "2",
            "paddingRight": "2",
            "paddingBottom": "2"
        },
        {
            "id": 34,
            "name": "Table Content",
            "bold": false,
            "italic": false,
            "underline": false,
            "strikethrough": false,
            "horizontalAlignment": "left",
            "verticalAlignment": "middle",
            "textColor": "#000000",
            "backgroundColor": "",
            "font": "helvetica",
            "fontSize": "9",
            "lineSpacing": 1,
            "borderColor": "#000000",
            "borderWidth": "1",
            "borderAll": false,
            "borderLeft": false,
            "borderTop": false,
            "borderRight": false,
            "borderBottom": false,
            "paddingLeft": "2",
            "paddingTop": "2",
            "paddingRight": "2",
            "paddingBottom": "2"
        },
        {
            "id": 35,
            "name": "Table Content Highlight",
            "bold": true,
            "italic": false,
            "underline": false,
            "strikethrough": false,
            "horizontalAlignment": "center",
            "verticalAlignment": "middle",
            "textColor": "#3d85c6",
            "backgroundColor": "",
            "font": "helvetica",
            "fontSize": "9",
            "lineSpacing": 1,
            "borderColor": "#000000",
            "borderWidth": "1",
            "borderAll": false,
            "borderLeft": false,
            "borderTop": false,
            "borderRight": false,
            "borderBottom": false,
            "paddingLeft": "2",
            "paddingTop": "2",
            "paddingRight": "2",
            "paddingBottom": "2"
        }
    ],
    "version": 3,
    "documentProperties": {
        "pageFormat": "A4",
        "pageWidth": "",
        "pageHeight": "",
        "unit": "mm",
        "orientation": "portrait",
        "contentHeight": "",
        "marginLeft": "10",
        "marginTop": "10",
        "marginRight": "10",
        "marginBottom": "10",
        "header": true,
        "headerSize": "80",
        "headerDisplay": "always",
        "footer": true,
        "footerSize": "30",
        "footerDisplay": "always",
        "patternLocale": "en",
        "patternCurrencySymbol": "$"
    }
}
"""

# TODO transform the SQL query into a Django ORM query
percentage_referrals_sql = """
SELECT CONCAT(HF."HFCode", ' - ', HF."HFName") HF, TotalClaim.TotalClaims, RefOP.TotalOP, RefIP.TotalIP
FROM (SELECT HF."HfID", HF."HFCode", HF."HFName"
      FROM "tblHF" HF
               INNER JOIN "uvwLocations" L ON L."LocationId" = HF."LocationId"
      WHERE HF."ValidityTo" Is NULL
        AND HF."HFLevel" IN ('D', 'C')
        AND (L."RegionId" = %(region_id)s OR %(region_id)s = 0 OR L."LocationId" IS NULL)
        AND (L."DistrictId" = %(district_id)s OR %(district_id)s = 0 OR L."DistrictId" IS NULL)
      ) HF
         LEFT OUTER JOIN (SELECT COUNT(1) TotalClaims, "HFID"
                          FROM "tblClaim"
                          WHERE "ValidityTo" Is NULL
                          AND "DateClaimed" BETWEEN %(date_start)s AND %(date_end)s
                          GROUP BY "HFID") TotalClaim ON HF."HfID" = TotalClaim."HFID"
         LEFT OUTER JOIN (SELECT I."HFID", COUNT(C."ClaimID") TotalOP
                          FROM "tblClaim" C
                                   INNER JOIN "tblInsuree" I ON C."InsureeID" = I."InsureeID"
                                   INNER JOIN "tblHF" HF ON C."HFID" = HF."HfID"
                                   INNER JOIN "uvwLocations" L ON L."LocationId" = HF."LocationId"
                          WHERE C."ValidityTo" Is NULL
                            AND I."ValidityTo" IS NULL
                            AND HF."ValidityTo" IS NULL
                            AND (C."DateTo" is null OR C."DateFrom" = C."DateTo")
                            AND HF."HfID" <> I."HFID"
                            AND C."VisitType" = N'R'
                            AND (L."RegionId" = %(region_id)s OR %(region_id)s = 0 OR L."LocationId" IS NULL)
                            AND (L."DistrictId" = %(district_id)s OR %(district_id)s = 0 OR L."DistrictId" IS NULL)
                            AND C."DateClaimed" BETWEEN %(date_start)s AND %(date_end)s
                          GROUP BY I."HFID") RefOP ON HF."HfID" = RefOP."HFID"
         LEFT OUTER JOIN (SELECT I."HFID", COUNT(C."ClaimID") TotalIP
                          FROM "tblClaim" C
                                   INNER JOIN "tblInsuree" I ON C."InsureeID" = I."InsureeID"
                                   INNER JOIN "tblHF" HF ON C."HFID" = HF."HfID"
                                   INNER JOIN "uvwLocations" L ON L."LocationId" = HF."LocationId"
                          WHERE C."ValidityTo" Is NULL
                            AND I."ValidityTo" IS NULL
                            AND HF."ValidityTo" IS NULL
                            AND C."DateTo" is not null and C."DateFrom" <> C."DateTo"
                            AND HF."HfID" <> I."HFID"
                            AND C."VisitType" = N'R'
                            AND (L."RegionId" = %(region_id)s OR %(region_id)s = 0 OR L."LocationId" IS NULL)
                            AND (L."DistrictId" = %(district_id)s OR %(district_id)s = 0 OR L."DistrictId" IS NULL)
                            AND C."DateClaimed" BETWEEN %(date_start)s AND %(date_end)s
                          GROUP BY I."HFID") RefIP ON HF."HfID" = RefIP."HFID"
"""


def claim_percentage_referrals_query(user, region_id=0, district_id=0, date_start="2019-01-01", date_end="2022-12-31",
                                     **kwargs):
    with connection.cursor() as cur:
        try:
            cur.execute(
                percentage_referrals_sql,
                {
                    "region_id": region_id,
                    "district_id": district_id,
                    "date_start": date_start,
                    "date_end": date_end,
                },
            )
            data = dictfetchall(cur)
            return {"data": data}
        except Exception as e:
            logger.exception("Error fetching claim percentage referrals query")
            raise e

    logger.error("Claim percentage referrals query arrived at end of function")
    return {"data": None}



from django.apps import AppConfig

MODULE_NAME = "claim"

DEFAULT_CFG = {
    "default_validations_disabled": False,
    "gql_query_claims_perms": ["111001"],
    "gql_query_claim_admins_perms": [],
    "gql_query_claim_officers_perms": [],
    "gql_query_claim_diagnosis_variance_only_on_existing": True,
    "gql_mutation_create_claims_perms": ["111002"],
    "gql_mutation_update_claims_perms": ["111010"],
    "gql_mutation_load_claims_perms": ["111005"],
    "gql_mutation_submit_claims_perms": ["111007"],
    "gql_mutation_select_claim_feedback_perms": ["111010"],
    "gql_mutation_bypass_claim_feedback_perms": ["111010"],
    "gql_mutation_skip_claim_feedback_perms": ["111010"],
    "gql_mutation_deliver_claim_feedback_perms": ["111009"],
    "gql_mutation_select_claim_review_perms": ["111010"],
    "gql_mutation_bypass_claim_review_perms": ["111010"],
    "gql_mutation_skip_claim_review_perms": ["111010"],
    "gql_mutation_deliver_claim_review_perms": ["111008"],
    "gql_mutation_process_claims_perms": ["111011"],
    "gql_mutation_restore_claims_perms": ["111012"],
    "gql_mutation_delete_claims_perms": ["111004"],
    "claim_print_perms": ["111006"],
    "claim_attachments_root_path": None,
    "claim_uspUpdateClaimFromPhone_intermediate_sets": 2,
    "autogenerated_claim_code_config": {'code_length': 8},
    "max_claim_length": 20,
    "claim_validation_multiple_services_explanation_required": True,
    "autogenerate_func": 'claim.utils.autogenerate_nepali_claim_code',
    "additional_diagnosis_number_allowed": 4,
    "claim_max_restore": None,
    "allowed_domains_attachments": [],
    "native_code_for_services": True
}


class ClaimConfig(AppConfig):
    name = MODULE_NAME

    default_validations_disabled = None
    gql_query_claims_perms = []
    gql_query_claim_admins_perms = []
    gql_query_claim_officers_perms = []
    gql_query_claim_diagnosis_variance_only_on_existing: None
    gql_mutation_create_claims_perms = []
    gql_mutation_update_claims_perms = []
    gql_mutation_load_claims_perms = []
    gql_mutation_submit_claims_perms = []
    gql_mutation_select_claim_feedback_perms = []
    gql_mutation_bypass_claim_feedback_perms = []
    gql_mutation_skip_claim_feedback_perms = []
    gql_mutation_deliver_claim_feedback_perms = []
    gql_mutation_select_claim_review_perms = []
    gql_mutation_bypass_claim_review_perms = []
    gql_mutation_skip_claim_review_perms = []
    gql_mutation_deliver_claim_review_perms = []
    gql_mutation_process_claims_perms = []
    gql_mutation_restore_claims_perms = []
    gql_mutation_delete_claims_perms = []
    claim_print_perms = []
    claim_attachments_root_path = None
    claim_uspUpdateClaimFromPhone_intermediate_sets = None
    autogenerated_claim_code_config = {}
    native_code_for_services = True
    # cannot be set in the model, since migration has to be able to handle all implementations
    max_claim_length = None
    claim_max_restore = None
    claim_validation_multiple_services_explanation_required = None
    # Provide absolute path to autogenerating function for claim code
    autogenerate_func = None
    additional_diagnosis_number_allowed = None  # Currently code supports 4 diagnoses maximum, going above will not work
    allowed_domains_attachments = None

    def __load_config(self, cfg):
        for field in cfg:
            if hasattr(ClaimConfig, field):
                setattr(ClaimConfig, field, cfg[field])

    def ready(self):
        from core.models import ModuleConfiguration
        cfg = ModuleConfiguration.get_or_default(MODULE_NAME, DEFAULT_CFG)
        self.__load_config(cfg)



# Generated by Django 3.2.19 on 2023-12-05 14:55

from django.db import migrations, models
import django.db.models.deletion
from django.conf import settings

class Migration(migrations.Migration):

    dependencies = [
        ('claim', '0025_auto_20231205_1455'),
    ]

    operations = [
        migrations.RunSQL(sql="""
            CREATE SEQUENCE IF NOT EXISTS tblfeedbackprompt_feedbackpromptid_seq;

            SELECT setval('tblfeedbackprompt_feedbackpromptid_seq', COALESCE((SELECT MAX("FeedbackPromptID") FROM "tblFeedbackPrompt"), 0) + 1, false);

            ALTER TABLE public."tblFeedbackPrompt"
            ALTER COLUMN "FeedbackPromptID" SET DEFAULT nextval('tblfeedbackprompt_feedbackpromptid_seq');

            ALTER SEQUENCE tblfeedbackprompt_feedbackpromptid_seq OWNED BY public."tblFeedbackPrompt"."FeedbackPromptID";
        """, reverse_sql=""),
        
    ] if not settings.MSSQL else []



# Generated by Django 3.2.16 on 2023-03-20 09:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('claim', '0013_feedbackprompt'),
    ]

    operations = [
        migrations.AlterField(
            model_name='claimadmin',
            name='code',
            field=models.CharField(max_length=50, blank=True, null=True),
        ),
    ]



from claim.reports import claim_percentage_referrals, claims_overview, claim_history, \
    claims_primary_operational_indicators
from claim.reports.claim_history import claim_history_query
from claim.reports.claim_percentage_referrals import claim_percentage_referrals_query
from claim.reports.claims_overview import claims_overview_query
from claim.reports.claims_primary_operational_indicators import claims_primary_operational_indicators_query

report_definitions = [
    {
        "name": "claim_percentage_referrals",
        "engine": 0,
        "default_report": claim_percentage_referrals.template,
        "description": "Percentage of referrals in claims",
        "module": "claim",
        "python_query": claim_percentage_referrals_query,
        "permission": ["131214"],
    },
    {
        "name": "claims_overview",
        "engine": 0,
        "default_report": claims_overview.template,
        "description": "Overview of the processing of claims",
        "module": "claim",
        "python_query": claims_overview_query,
        "permission": ["131213"],
    },
    {
        "name": "claim_history",
        "engine": 0,
        "default_report": claim_history.template,
        "description": "Claim history",
        "module": "claim",
        "python_query": claim_history_query,
        "permission": ["131223"],
    },
    {
        "name": "claims_primary_operational_indicators",
        "engine": 0,
        "default_report": claims_primary_operational_indicators.template,
        "description": "Claims Primary operational indicators",
        "module": "claim",
        "python_query": claims_primary_operational_indicators_query,
        "permission": ["131202"],
    },
]



import uuid

from claim_batch import models as claim_batch_models
from core import fields, TimeUtils
from core import models as core_models
from django import dispatch
from django.conf import settings
from django.db import models
from graphql import ResolveInfo
from insuree import models as insuree_models
from location import models as location_models
from location.models import LocationManager
from medical import models as medical_models
from policy import models as policy_models
from product import models as product_models
from django.apps import apps
from django.utils import timezone as django_tz

core_config = apps.get_app_config('core')


class ClaimAdmin(core_models.VersionedModel):
    id = models.AutoField(db_column='ClaimAdminId', primary_key=True)
    uuid = models.CharField(db_column='ClaimAdminUUID', max_length=36, default=uuid.uuid4, unique=True)

    code = models.CharField(db_column='ClaimAdminCode', max_length=50,
                            blank=True, null=True)
    last_name = models.CharField(db_column='LastName', max_length=100, blank=True, null=True)
    other_names = models.CharField(db_column='OtherNames', max_length=100, blank=True, null=True)
    dob = models.DateField(db_column='DOB', blank=True, null=True)
    email_id = models.CharField(db_column='EmailId', max_length=200, blank=True, null=True)
    phone = models.CharField(db_column='Phone', max_length=50, blank=True, null=True)
    health_facility = models.ForeignKey(
        location_models.HealthFacility, models.DO_NOTHING, db_column='HFId', blank=True, null=True)
    has_login = models.BooleanField(db_column='HasLogin', blank=True, null=True)

    audit_user_id = models.IntegerField(db_column='AuditUserId', blank=True, null=True)
    # row_id = models.BinaryField(db_column='RowId', blank=True, null=True)

    def __str__(self):
        return self.code + " " + self.last_name + " " + self.other_names

    @classmethod
    def get_queryset(cls, queryset, user):
        queryset = cls.filter_queryset(queryset)
        # GraphQL calls with an info object while Rest calls with the user itself
        if isinstance(user, ResolveInfo):
            user = user.context.user
        if settings.ROW_SECURITY and user.is_anonymous:
            return queryset.filter(id=-1)
        if settings.ROW_SECURITY:
            from location.schema import LocationManager
            queryset = LocationManager().build_user_location_filter_query(
                user._u, prefix='health_facility__location', queryset=queryset, loc_types=['D'])
        return queryset

    @property
    def id_for_audit(self):
        return self.audit_user_id

    @property
    def username(self):
        return self.code

    def get_username(self):
        return self.code

    @property
    def is_staff(self):
        return False

    @property
    def is_superuser(self):
        return False

    def set_password(self, raw_password):
        raise NotImplementedError("Shouldn't set a password on an Officer")

    def check_password(self, raw_password):
        return False

    @property
    def officer_allowed_locations(self):
        """
        Returns uuid of all locations allowed for given officerLocationManager
        """
        district = self.health_facility.location
        all_allowed_uuids = [district.parent.uuid, district.uuid]
        child_locations = location_models.Location.objects.filter(parent=district).values_list('uuid', flat=True)
        while child_locations:
            all_allowed_uuids.extend(child_locations)
            child_locations = location_models.Location.objects\
                .filter(parent__uuid__in=child_locations)\
                .values_list('uuid', flat=True)

        return location_models.Location.objects.filter(uuid__in=all_allowed_uuids)

    class Meta:
        managed = True
        db_table = 'tblClaimAdmin'


class Feedback(core_models.VersionedModel):
    id = models.AutoField(db_column='FeedbackID', primary_key=True)
    uuid = models.CharField(db_column='FeedbackUUID', max_length=36, default=uuid.uuid4, unique=True)
    claim = models.OneToOneField(
        "Claim", models.DO_NOTHING, db_column='ClaimID', blank=True, null=True, related_name="+")
    care_rendered = models.BooleanField(db_column='CareRendered', blank=True, null=True)
    payment_asked = models.BooleanField(db_column='PaymentAsked', blank=True, null=True)
    drug_prescribed = models.BooleanField(db_column='DrugPrescribed', blank=True, null=True)
    drug_received = models.BooleanField(db_column='DrugReceived', blank=True, null=True)
    asessment = models.SmallIntegerField(db_column='Asessment', blank=True, null=True)
    # No FK in database (so value may not be an existing officer.id !)
    officer_id = models.IntegerField(db_column='CHFOfficerCode', blank=True, null=True)
    feedback_date = fields.DateTimeField(db_column='FeedbackDate', blank=True, null=True)
    audit_user_id = models.IntegerField(db_column='AuditUserID')

    class Meta:
        managed = True
        db_table = 'tblFeedback'

    @classmethod
    def get_queryset(cls, queryset, user):
        queryset = cls.filter_queryset(queryset)
        # GraphQL calls with an info object while Rest calls with the user itself
        if isinstance(user, ResolveInfo):
            user = user.context.user
        if settings.ROW_SECURITY and user.is_anonymous:
            return queryset.filter(id=-1)
        if settings.ROW_SECURITY:
            queryset = LocationManager().build_user_location_filter_query(
                user._u, prefix='health_facility__location', queryset=queryset, loc_types=['D'])
        return queryset



signal_claim_rejection = dispatch.Signal(["claim"])


class Claim(core_models.VersionedModel, core_models.ExtendableModel):
    id = models.AutoField(db_column='ClaimID', primary_key=True)
    uuid = models.CharField(db_column='ClaimUUID',
                            max_length=36, default=uuid.uuid4, unique=True)
    category = models.CharField(
        db_column='ClaimCategory', max_length=1, blank=True, null=True)
    insuree = models.ForeignKey(
        insuree_models.Insuree, models.DO_NOTHING, db_column='InsureeID')
    # do not change max_length value - use setting from apps.py
    code = models.CharField(db_column='ClaimCode', max_length=50)
    date_from = fields.DateField(db_column='DateFrom')
    date_to = fields.DateField(db_column='DateTo', blank=True, null=True)
    status = models.SmallIntegerField(db_column='ClaimStatus')
    restore = models.ForeignKey('self', db_column='RestoredClaim', on_delete=models.DO_NOTHING, blank=True, null=True)
    adjuster = models.ForeignKey(
        core_models.InteractiveUser, models.DO_NOTHING,
        db_column='Adjuster', blank=True, null=True)
    adjustment = models.TextField(
        db_column='Adjustment', blank=True, null=True)
    claimed = models.DecimalField(
        db_column='Claimed',
        max_digits=18, decimal_places=2, blank=True, null=True)
    approved = models.DecimalField(
        db_column='Approved',
        max_digits=18, decimal_places=2, blank=True, null=True)
    reinsured = models.DecimalField(
        db_column='Reinsured',
        max_digits=18, decimal_places=2, blank=True, null=True)
    valuated = models.DecimalField(
        db_column='Valuated', max_digits=18, decimal_places=2, blank=True, null=True)
    date_claimed = fields.DateField(db_column='DateClaimed')
    date_processed = fields.DateField(
        db_column='DateProcessed', blank=True, null=True)
    # Django uses the feedback_id column to create the feedback column, which conflicts with the boolean field
    feedback_available = models.BooleanField(
        db_column='Feedback', default=False)
    feedback = models.OneToOneField(
        Feedback, models.DO_NOTHING,
        db_column='FeedbackID', blank=True, null=True, related_name="+")
    explanation = models.TextField(
        db_column='Explanation', blank=True, null=True)
    feedback_status = models.SmallIntegerField(
        db_column='FeedbackStatus', blank=True, null=True, default=1)
    review_status = models.SmallIntegerField(
        db_column='ReviewStatus', blank=True, null=True, default=1)
    approval_status = models.SmallIntegerField(
        db_column='ApprovalStatus', blank=True, null=True, default=1)
    rejection_reason = models.SmallIntegerField(
        db_column='RejectionReason', blank=True, null=True, default=0)

    batch_run = models.ForeignKey(claim_batch_models.BatchRun,
                                  models.DO_NOTHING, db_column='RunID', blank=True, null=True)
    audit_user_id = models.IntegerField(db_column='AuditUserID')
    validity_from_review = fields.DateTimeField(
        db_column='ValidityFromReview', blank=True, null=True)
    validity_to_review = fields.DateTimeField(
        db_column='ValidityToReview', blank=True, null=True)

    health_facility = models.ForeignKey(
        location_models.HealthFacility, models.DO_NOTHING, db_column='HFID')

    submit_stamp = fields.DateTimeField(
        db_column='SubmitStamp', blank=True, null=True)
    process_stamp = fields.DateTimeField(
        db_column='ProcessStamp', blank=True, null=True)
    remunerated = models.DecimalField(
        db_column='Remunerated', max_digits=18, decimal_places=2, blank=True, null=True)
    guarantee_id = models.CharField(
        db_column='GuaranteeId', max_length=50, blank=True, null=True)
    admin = models.ForeignKey(
        ClaimAdmin, models.DO_NOTHING, db_column='ClaimAdminId',
        blank=True, null=True)
    refer_from = models.ForeignKey(
        location_models.HealthFacility, models.DO_NOTHING, related_name='referFromHF',
        db_column='ReferFrom', blank=True, null=True)
    refer_to = models.ForeignKey(
        location_models.HealthFacility, models.DO_NOTHING, related_name='referToHF',
        db_column='ReferTo', blank=True, null=True)
    icd = models.ForeignKey(
        medical_models.Diagnosis, models.DO_NOTHING, db_column='ICDID',
        related_name="claim_icds")
    icd_1 = models.ForeignKey(
        medical_models.Diagnosis, models.DO_NOTHING, db_column='ICDID1',
        related_name="claim_icd1s",
        blank=True, null=True)
    icd_2 = models.ForeignKey(
        medical_models.Diagnosis, models.DO_NOTHING, db_column='ICDID2',
        related_name="claim_icd2s",
        blank=True, null=True)
    icd_3 = models.ForeignKey(
        medical_models.Diagnosis, models.DO_NOTHING, db_column='ICDID3',
        related_name="claim_icd3s",
        blank=True, null=True)
    icd_4 = models.ForeignKey(
        medical_models.Diagnosis, models.DO_NOTHING, db_column='ICDID4',
        related_name="claim_icd4s",
        blank=True, null=True)

    visit_type = models.CharField(
        db_column='VisitType', max_length=1, blank=True, null=True)
    audit_user_id_review = models.IntegerField(
        db_column='AuditUserIDReview', blank=True, null=True)
    audit_user_id_submit = models.IntegerField(
        db_column='AuditUserIDSubmit', blank=True, null=True)
    audit_user_id_process = models.IntegerField(
        db_column='AuditUserIDProcess', blank=True, null=True)
    care_type = models.CharField(db_column='CareType', max_length=4, blank=True, null=True)

    # row_id = models.BinaryField(db_column='RowID', blank=True, null=True)

    class Meta:
        managed = True
        db_table = 'tblClaim'

    STATUS_REJECTED = 1
    STATUS_ENTERED = 2
    STATUS_CHECKED = 4
    STATUS_PROCESSED = 8
    STATUS_VALUATED = 16

    FEEDBACK_IDLE = 1
    FEEDBACK_NOT_SELECTED = 2
    FEEDBACK_SELECTED = 4
    FEEDBACK_DELIVERED = 8
    FEEDBACK_BYPASSED = 16

    REVIEW_IDLE = 1
    REVIEW_NOT_SELECTED = 2
    REVIEW_SELECTED = 4
    REVIEW_DELIVERED = 8
    REVIEW_BYPASSED = 16

    def reject(self, rejection_code):
        updated_items = self.items.filter(validity_to__isnull=True).update(
            rejection_reason=rejection_code)
        updated_services = self.services.filter(
            validity_to__isnull=True).update(rejection_reason=rejection_code)
        signal_claim_rejection.send(sender=self.__class__, claim=self)
        return updated_items + updated_services

    def save_history(self, **kwargs):
        prev_id = super(Claim, self).save_history()
        if prev_id:
            prev_items = []
            for item in self.items.all():
                prev_items.append(item.save_history())
            ClaimItem.objects.filter(
                id__in=prev_items).update(claim_id=prev_id)
            prev_services = []
            for service in self.services.all():
                prev_services.append(service.save_history())
            ClaimService.objects.filter(
                id__in=prev_services).update(claim_id=prev_id)
        return prev_id

    @classmethod
    def get_queryset(cls, queryset, user):
        queryset = Claim.filter_queryset(queryset)
        # GraphQL calls with an info object while Rest calls with the user itself
        if isinstance(user, ResolveInfo):
            user = user.context.user
        if settings.ROW_SECURITY and user.is_anonymous:
            return queryset.filter(id=-1)
        if settings.ROW_SECURITY:
            # TechnicalUsers don't have health_facility_id attribute
            if hasattr(user._u, 'health_facility_id') and user._u.health_facility_id:
                queryset = queryset.filter(
                    health_facility_id=user._u.health_facility_id
                )
            else:
                if not isinstance(user._u, core_models.TechnicalUser):
                    queryset = LocationManager().build_user_location_filter_query(
                        user._u, prefix='health_facility__location', queryset=queryset, loc_types=['D'])
        return queryset


class FeedbackPrompt(core_models.VersionedModel):
    id = models.AutoField(db_column='FeedbackPromptID', primary_key=True)
    feedback_prompt_date = fields.DateField(db_column='FeedbackPromptDate', blank=True, null=True)
    claim = models.ForeignKey(
        Claim, models.DO_NOTHING, db_column='ClaimID', blank=True, null=True, related_name="+")
    officer = models.ForeignKey(
        core_models.Officer, models.DO_NOTHING, db_column="OfficerID", blank=True, null=True)
    phone_number = models.CharField(db_column='PhoneNumber', max_length=50, blank=True, null=True)
    sms_status = models.IntegerField(db_column='SMSStatus', blank=True, null=True)
    validity_from = fields.DateTimeField(db_column='ValidityFrom', blank=True, null=True)
    validity_to = fields.DateTimeField(db_column='ValidityTo', blank=True, null=True)
    legacy_id = models.IntegerField(db_column='LegacyID', blank=True, null=True)
    audit_user_id = models.IntegerField(db_column='AuditUserID', blank=True, null=True)

    class Meta:
        managed = True
        db_table = 'tblFeedbackPrompt'

    @classmethod
    def get_queryset(cls, queryset, user):
        queryset = cls.filter_queryset(queryset)
        # GraphQL calls with an info object while Rest calls with the user itself
        if isinstance(user, ResolveInfo):
            user = user.context.user
        if settings.ROW_SECURITY and user.is_anonymous:
            return queryset.filter(id=-1)
        if settings.ROW_SECURITY:
            queryset = LocationManager().build_user_location_filter_query(
                user._u, prefix='health_facility__location', queryset=queryset, loc_types=['D'])

        return queryset



class ClaimAttachmentsCount(models.Model):
    claim = models.OneToOneField(Claim, primary_key=True, related_name='attachments_count', on_delete=models.DO_NOTHING)
    value = models.IntegerField(db_column='attachments_count')

    class Meta:
        managed = True
        db_table = 'claim_ClaimAttachmentsCountView'


class ClaimMutation(core_models.UUIDModel):
    claim = models.ForeignKey(Claim, models.DO_NOTHING,
                              related_name='mutations')
    mutation = models.ForeignKey(
        core_models.MutationLog, models.DO_NOTHING, related_name='claims')

    class Meta:
        managed = True
        db_table = "claim_ClaimMutation"


class ClaimDetailManager(models.Manager):

    def filter(self, *args, **kwargs):
        keys = [x for x in kwargs if "itemsvc" in x]
        for key in keys:
            new_key = key.replace("itemsvc", self.model.model_prefix)
            kwargs[new_key] = kwargs.pop(key)
        return super(ClaimDetailManager, self).filter(*args, **kwargs)


class ClaimDetail:
    STATUS_PASSED = 1
    STATUS_REJECTED = 2

    objects = ClaimDetailManager()

    @property
    def itemsvc(self):
        if hasattr(self, "item"):
            return self.item
        elif hasattr(self, "service"):
            return self.service
        else:
            raise Exception("ClaimDetail has neither item nor service")

    class Meta:
        abstract = True


class ClaimItem(core_models.VersionedModel, ClaimDetail, core_models.ExtendableModel):
    model_prefix = "item"
    id = models.AutoField(db_column='ClaimItemID', primary_key=True)
    claim = models.ForeignKey(Claim, models.DO_NOTHING,
                              db_column='ClaimID', related_name='items')
    item = models.ForeignKey(
        medical_models.Item, models.DO_NOTHING, db_column='ItemID')
    product = models.ForeignKey(product_models.Product,
                                models.DO_NOTHING, db_column='ProdID',
                                blank=True, null=True,
                                related_name="claim_items")
    status = models.SmallIntegerField(db_column='ClaimItemStatus')
    availability = models.BooleanField(db_column='Availability')
    qty_provided = models.DecimalField(
        db_column='QtyProvided', max_digits=18, decimal_places=2)
    qty_approved = models.DecimalField(
        db_column='QtyApproved', max_digits=18, decimal_places=2, blank=True, null=True)
    price_asked = models.DecimalField(
        db_column='PriceAsked', max_digits=18, decimal_places=2)
    price_adjusted = models.DecimalField(
        db_column='PriceAdjusted', max_digits=18, decimal_places=2, blank=True, null=True)
    price_approved = models.DecimalField(
        db_column='PriceApproved', max_digits=18, decimal_places=2, blank=True, null=True)
    price_valuated = models.DecimalField(
        db_column='PriceValuated', max_digits=18, decimal_places=2, blank=True, null=True)
    explanation = models.TextField(
        db_column='Explanation', blank=True, null=True)
    justification = models.TextField(
        db_column='Justification', blank=True, null=True)
    rejection_reason = models.SmallIntegerField(
        db_column='RejectionReason', blank=True, null=True)
    audit_user_id = models.IntegerField(db_column='AuditUserID')
    validity_from_review = fields.DateTimeField(
        db_column='ValidityFromReview', blank=True, null=True)
    validity_to_review = fields.DateTimeField(
        db_column='ValidityToReview', blank=True, null=True)
    audit_user_id_review = models.IntegerField(
        db_column='AuditUserIDReview', blank=True, null=True)
    limitation_value = models.DecimalField(
        db_column='LimitationValue', max_digits=18, decimal_places=2, blank=True, null=True)
    limitation = models.CharField(
        db_column='Limitation', max_length=1, blank=True, null=True)
    policy = models.ForeignKey(
        policy_models.Policy, models.DO_NOTHING, db_column='PolicyID', blank=True, null=True)
    remunerated_amount = models.DecimalField(
        db_column='RemuneratedAmount', max_digits=18, decimal_places=2, blank=True, null=True)
    deductable_amount = models.DecimalField(
        db_column='DeductableAmount', max_digits=18, decimal_places=2, blank=True, null=True)
    exceed_ceiling_amount = models.DecimalField(
        db_column='ExceedCeilingAmount', max_digits=18, decimal_places=2, blank=True, null=True)
    price_origin = models.CharField(
        db_column='PriceOrigin', max_length=1, blank=True, null=True)
    exceed_ceiling_amount_category = models.DecimalField(
        db_column='ExceedCeilingAmountCategory', max_digits=18, decimal_places=2, blank=True, null=True)
    objects = ClaimDetailManager()

    class Meta:
        managed = True
        db_table = 'tblClaimItems'


class GeneralClaimAttachmentType(models.TextChoices):
    URL = "URL"
    FILE = "FILE"


class ClaimAttachmentType(core_models.VersionedModel):
    id = models.SmallIntegerField(db_column='ClaimAttachmentTypeId', primary_key=True)
    claim_attachment_type = models.CharField(db_column='ClaimAttachmentType', max_length=50)
    is_autogenerated = models.BooleanField(default=False)
    claim_general_type = models.CharField(max_length=10, default=GeneralClaimAttachmentType.FILE,
                                          choices=GeneralClaimAttachmentType.choices)

    class Meta:
        managed = True
        db_table = 'claim_ClaimAttachment_ClaimAttachmentType'


class ClaimAttachment(core_models.UUIDModel, core_models.UUIDVersionedModel):
    claim = models.ForeignKey(
        Claim, models.DO_NOTHING, related_name='attachments')
    general_type = models.CharField(max_length=4, choices=GeneralClaimAttachmentType.choices,
                                    default=GeneralClaimAttachmentType.FILE)
    type = models.TextField(blank=True, null=True)
    predefined_type = models.ForeignKey(ClaimAttachmentType, models.DO_NOTHING, related_name='type_dropdown', null=True,
                                        blank=True)
    title = models.TextField(blank=True, null=True)
    date = fields.DateField(blank=True, default=TimeUtils.now)
    filename = models.TextField(blank=True, null=True)
    mime = models.TextField(blank=True, null=True)
    # this is not needed at the moment, but we want to move attachment to core
    # in that case module information is needed, and we want to avoid writing additional migration
    module = models.TextField(blank=False, null=True)
    # frontend contributions may lead to externalized (nas) storage for documents
    url = models.TextField(blank=True, null=True)
    # Support of BinaryField is database-related: prefer to stick to b64-encoded
    document = models.TextField(blank=True, null=True)

    class Meta:
        managed = True
        db_table = "claim_ClaimAttachment"


class ClaimService(core_models.VersionedModel, ClaimDetail, core_models.ExtendableModel):
    model_prefix = "service"
    id = models.AutoField(db_column='ClaimServiceID', primary_key=True)
    claim = models.ForeignKey(
        Claim, models.DO_NOTHING, db_column='ClaimID', related_name='services')
    service = models.ForeignKey(
        medical_models.Service, models.DO_NOTHING, db_column='ServiceID')
    product = models.ForeignKey(product_models.Product,
                                models.DO_NOTHING, db_column='ProdID',
                                blank=True, null=True,
                                related_name="claim_services")
    status = models.SmallIntegerField(db_column='ClaimServiceStatus')
    qty_provided = models.DecimalField(
        db_column='QtyProvided', max_digits=18, decimal_places=2)
    qty_approved = models.DecimalField(
        db_column='QtyApproved', max_digits=18, decimal_places=2, blank=True, null=True)
    price_asked = models.DecimalField(
        db_column='PriceAsked', max_digits=18, decimal_places=2)
    price_adjusted = models.DecimalField(
        db_column='PriceAdjusted', max_digits=18, decimal_places=2, blank=True, null=True)
    price_approved = models.DecimalField(
        db_column='PriceApproved', max_digits=18, decimal_places=2, blank=True, null=True)
    price_valuated = models.DecimalField(
        db_column='PriceValuated', max_digits=18, decimal_places=2, blank=True, null=True)
    explanation = models.TextField(
        db_column='Explanation', blank=True, null=True)
    justification = models.TextField(
        db_column='Justification', blank=True, null=True)
    rejection_reason = models.SmallIntegerField(
        db_column='RejectionReason', blank=True, null=True)
    audit_user_id = models.IntegerField(db_column='AuditUserID')
    validity_from_review = fields.DateTimeField(
        db_column='ValidityFromReview', blank=True, null=True)
    validity_to_review = fields.DateTimeField(
        db_column='ValidityToReview', blank=True, null=True)
    audit_user_id_review = models.IntegerField(
        db_column='AuditUserIDReview', blank=True, null=True)
    limitation_value = models.DecimalField(
        db_column='LimitationValue', max_digits=18, decimal_places=2, blank=True, null=True)
    limitation = models.CharField(
        db_column='Limitation', max_length=1, blank=True, null=True)
    policy = models.ForeignKey(
        policy_models.Policy, models.DO_NOTHING, db_column='PolicyID', blank=True, null=True)
    remunerated_amount = models.DecimalField(
        db_column='RemuneratedAmount', max_digits=18, decimal_places=2, blank=True, null=True)
    deductable_amount = models.DecimalField(
        db_column='DeductableAmount', max_digits=18, decimal_places=2, blank=True, null=True)
    exceed_ceiling_amount = models.DecimalField(
        db_column='ExceedCeilingAmount', max_digits=18, decimal_places=2, blank=True, null=True)
    price_origin = models.CharField(
        db_column='PriceOrigin', max_length=1, blank=True, null=True)
    exceed_ceiling_amount_category = models.DecimalField(
        db_column='ExceedCeilingAmountCategory', max_digits=18, decimal_places=2, blank=True, null=True)
    objects = ClaimDetailManager()

    class Meta:
        managed = True
        db_table = 'tblClaimServices'

class ClaimServiceItem(models.Model):
    id = models.AutoField(primary_key=True, db_column='idCsi')
    item = models.ForeignKey(medical_models.Item, models.DO_NOTHING, db_column='ItemID', related_name="service_items")                           
    claim_service = models.ForeignKey( ClaimService,
                                          models.DO_NOTHING, db_column="ClaimServiceID",related_name='items')
    qty_provided = models.IntegerField(db_column="qty_provided",
                                      blank=True, null=True)
    qty_displayed = models.IntegerField(db_column="qty_displayed",
                                      blank=True, null=True)
    created_date = models.DateTimeField(db_column="created_date", default=django_tz.now,
                                   blank=True, null=True)
    price_asked = models.DecimalField(db_column="price",
                                   max_digits=18, decimal_places=2, blank=True, null=True)
                                   
    class Meta:
        managed = True
        db_table = 'tblClaimServicesItems'

class ClaimServiceService(models.Model):
    id = models.AutoField(primary_key=True, db_column='idCss')
    service = models.ForeignKey(medical_models.Service, models.DO_NOTHING,
                              db_column='ServiceId', related_name='service_services')
    claim_service = models.ForeignKey( ClaimService,
                                          models.DO_NOTHING, db_column="claimServiceID", related_name='services')
    qty_provided = models.IntegerField(db_column="qty_provided",
                                      blank=True, null=True)
    qty_displayed = models.IntegerField(db_column="qty_displayed",
                                      blank=True, null=True)
    created_date = models.DateTimeField(db_column="created_date", default=django_tz.now,
                                   blank=True, null=True)
    price_asked = models.DecimalField(db_column="price",
                                   max_digits=18, decimal_places=2, blank=True, null=True)
    
    class Meta:
        managed = True
        db_table = 'tblClaimServicesService'

class ClaimDedRem(core_models.VersionedModel):
    id = models.AutoField(db_column='ExpenditureID', primary_key=True)

    policy = models.ForeignKey('policy.Policy', models.DO_NOTHING, db_column='PolicyID', blank=True, null=True,
                               related_name='claim_ded_rems')
    insuree = models.ForeignKey('insuree.Insuree', models.DO_NOTHING, db_column='InsureeID', blank=True, null=True,
                                related_name='claim_ded_rems')
    claim = models.ForeignKey(to=Claim, db_column='ClaimID', db_index=True, related_name="dedrems",
                              on_delete=models.DO_NOTHING)
    ded_g = models.DecimalField(db_column='DedG', max_digits=18, decimal_places=2, blank=True, null=True)
    ded_op = models.DecimalField(db_column='DedOP', max_digits=18, decimal_places=2, blank=True, null=True)
    ded_ip = models.DecimalField(db_column='DedIP', max_digits=18, decimal_places=2, blank=True, null=True)
    rem_g = models.DecimalField(db_column='RemG', max_digits=18, decimal_places=2, blank=True, null=True)
    rem_op = models.DecimalField(db_column='RemOP', max_digits=18, decimal_places=2, blank=True, null=True)
    rem_ip = models.DecimalField(db_column='RemIP', max_digits=18, decimal_places=2, blank=True, null=True)
    rem_consult = models.DecimalField(db_column='RemConsult', max_digits=18, decimal_places=2, blank=True, null=True)
    rem_surgery = models.DecimalField(db_column='RemSurgery', max_digits=18, decimal_places=2, blank=True, null=True)
    rem_delivery = models.DecimalField(db_column='RemDelivery', max_digits=18, decimal_places=2, blank=True, null=True)
    rem_hospitalization = models.DecimalField(db_column='RemHospitalization', max_digits=18, decimal_places=2,
                                              blank=True, null=True)
    rem_antenatal = models.DecimalField(db_column='RemAntenatal', max_digits=18, decimal_places=2,
                                        blank=True, null=True)

    audit_user_id = models.IntegerField(db_column='AuditUserID')

    class Meta:
        managed = True
        db_table = 'tblClaimDedRem'



import random
import string

from claim.test_helpers import create_test_claim_admin
from django.core.management.base import BaseCommand
from faker import Faker


class Command(BaseCommand):
    help = "This command will generate test Claims Administrators. It is intended to simulate larger" \
           "databases for performance testing. Some locations have almost 1000 of them."
    insurees = None
    services = None
    items = None
    claim_admins = None
    hfs = None

    def add_arguments(self, parser):
        parser.add_argument("nb_admins", nargs=1, type=int)
        parser.add_argument(
            '--verbose',
            action='store_true',
            dest='verbose',
            help='Be verbose about what it is doing',
        )
        parser.add_argument(
            '--locale',
            default="fr_FR",
            help="Used to adapt the fake names generation to the locale, using Faker, by default fr_FR",
        )

    def handle(self, *args, **options):

        nb_admins = options["nb_admins"][0]
        verbose = options["verbose"]
        fake = Faker(options["locale"])
        for admin_num in range(1, nb_admins + 1):
            hf = self.get_random_hf()
            claim_admin = create_test_claim_admin({
                "code": ''.join(random.choices(string.ascii_uppercase + string.digits, k=8)),
                "last_name": fake.last_name(),
                "other_names": fake.first_name(),
                "email_id": fake.ascii_email(),
                "phone": "+" + fake.msisdn(),
                "health_facility_id": hf})
            if verbose:
                logger.debug(f"{admin_num} created claim admin {claim_admin} for HF {hf} \
                    with code {claim_admin.code}")

    def get_random_hf(self):
        if not self.hfs:
            from location.models import HealthFacility
            self.hfs = HealthFacility.objects.filter(validity_to__isnull=True).values_list("pk", flat=True)
        return random.choice(self.hfs)



# Generated by Django 2.1.11 on 2019-10-28 06:40

import core.fields
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('claim', '0003_claimofficer'),
    ]

    operations = [
        migrations.CreateModel(
            name='ClaimAttachment',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('legacy_id', models.UUIDField(blank=True, db_column='LegacyID', null=True)),
                ('validity_from', core.fields.DateTimeField(db_column='ValidityFrom')),
                ('validity_to', core.fields.DateTimeField(blank=True, db_column='ValidityTo', null=True)),
                ('type', models.TextField(blank=True, null=True)),
                ('title', models.TextField(blank=True, null=True)),
                ('date', core.fields.DateField(blank=True, null=True)),
                ('filename', models.TextField(blank=True, null=True)),
                ('mime', models.TextField(blank=True, null=True)),
                ('document', models.TextField(blank=True, null=True)),
                ('claim', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, related_name='attachments', to='claim.Claim')),
            ],
            options={
                'db_table': 'claim_ClaimAttachment',
                'managed': True,
            },
        ),
    ]
